from redis.asyncio import Redis

from zayt.conf.settings import Settings


def make_service(name: str):
    async def redis_service(settings: Settings) -> Redis:
        redis_settings = dict(settings.redis[name])

        if url := redis_settings.pop("url", None):
            redis = Redis.from_url(url, **redis_settings)
        else:
            redis = Redis(**redis_settings)

        await redis.initialize()
        yield redis
        await redis.aclose()

    return redis_service



from importlib.util import find_spec

from zayt.conf.settings import Settings
from zayt.di.container import Container
from zayt.web.lifecycle.decorator import startup

from .service import make_service


@startup
def init(container: Container, settings: Settings):
    if find_spec("redis") is None:
        raise ModuleNotFoundError("Missing 'redis'. Install 'zayt' with 'redis' extra.")

    for name in settings.redis:
        assert isinstance(name, str)
        name = name.strip()
        container.register(make_service(name), name=name if name else None)



from http import HTTPMethod

import pytest

from zayt.web.routing.exception import DuplicateRouteError
from zayt.web.routing.router import Router


def test_duplicate_route_should_raise_error():
    async def route1(_request, _response):
        pass

    async def route2(_request, _response):
        pass

    router = Router()

    with pytest.raises(DuplicateRouteError):
        router.route(route1, HTTPMethod.GET, "/")
        router.route(route2, HTTPMethod.GET, "/")



import asyncio
from typing import Annotated

from zayt.di.container import Container
from zayt.di.inject import Inject


class Service1:
    pass


class Service2:
    dep1: Annotated[Service1, Inject]


class Service3:
    dep1: Annotated[Service1, Inject]
    dep2: Annotated[Service2, Inject]

class Service4:
    dep1: Annotated[Service1, Inject]
    dep2: Annotated[Service2, Inject]
    dep3: Annotated[Service3, Inject]


class Service5:
    dep1: Annotated[Service1, Inject]
    dep2: Annotated[Service2, Inject]
    dep3: Annotated[Service3, Inject]
    dep4: Annotated[Service4, Inject]
    dep6: Annotated["Service6", Inject]


class Service6:
    dep1: Annotated[Service1, Inject]
    dep2: Annotated[Service2, Inject]
    dep3: Annotated[Service3, Inject]
    dep4: Annotated[Service4, Inject]
    dep5: Annotated[Service5, Inject]


async def test_get_service(ioc: Container):
    for svc in [Service1, Service2, Service3, Service4]:
        ioc.register(svc)

    instance = await ioc.get(Service4)
    assert isinstance(instance, Service4)


async def test_dependency_loop(ioc: Container):
    for svc in [Service1, Service2, Service3, Service4, Service5, Service6]:
        ioc.register(svc)

    instance5 = await ioc.get(Service5)
    instance6 = await ioc.get(Service6)

    assert isinstance(instance5, Service5)
    assert isinstance(instance6, Service6)

    assert instance5.dep6 is instance6
    assert instance6.dep5 is instance5


async def test_get_simultaneously(ioc: Container):
    for svc in [Service1, Service2, Service3, Service4, Service5, Service6]:
        ioc.register(svc)

    (
        instance1,
        instance2,
        instance3,
        instance4,
        instance5,
        instance6,
    ) = await asyncio.gather(
        ioc.get(Service1),
        ioc.get(Service2),
        ioc.get(Service3),
        ioc.get(Service4),
        ioc.get(Service5),
        ioc.get(Service6),
    )

    assert isinstance(instance1, Service1)
    assert isinstance(instance2, Service2)
    assert isinstance(instance3, Service3)
    assert isinstance(instance4, Service4)
    assert isinstance(instance5, Service5)
    assert isinstance(instance6, Service6)

    assert instance5.dep6 is instance6
    assert instance6.dep5 is instance5



import asyncio
import multiprocessing as mp
from pathlib import Path

from httpx import ASGITransport, AsyncClient

from .util import with_app


@with_app
async def _run(app, message):
    async with AsyncClient(
        transport=ASGITransport(app=app), base_url="http://server"
    ) as client:
        response = await client.get("/")
        assert response.text == message


def run(profile, message):
    import os

    os.environ["ZAYT_PROFILE"] = profile
    return asyncio.run(_run(message))


def test_example(monkeypatch):
    path = Path(__file__).parents[2] / "examples" / "settings"
    monkeypatch.syspath_prepend(path)
    monkeypatch.chdir(path)

    with mp.Pool(processes=1) as pool:
        pool.apply(run, ["", "Hello, World!"])
        pool.apply(run, ["dev", "Hello, dev World!"])



import pytest
from sqlalchemy import String, select
from sqlalchemy.ext.asyncio import AsyncSession, create_async_engine
from sqlalchemy.orm import DeclarativeBase, Mapped, mapped_column

from zayt.conf.defaults import default_settings
from zayt.conf.settings import Settings
from zayt.ext.sqlalchemy.service import sessionmaker_service


class BaseA(DeclarativeBase):
    pass


class ModelA(BaseA):
    __tablename__ = "model"
    id: Mapped[int] = mapped_column(primary_key=True)
    name: Mapped[str] = mapped_column(String(100))


class BaseB(DeclarativeBase):
    pass


class ModelB(BaseB):
    __tablename__ = "model"
    id: Mapped[int] = mapped_column(primary_key=True)
    name: Mapped[str] = mapped_column(String(100))


async def test_session_options():
    settings = Settings(
        default_settings
        | {
            "sqlalchemy": {
                "connections": {
                    "default": {
                        "url": "sqlite+aiosqlite:///:memory:",
                    },
                },
                "session": {
                    "options": {
                        "info": {"framework": "zayt"},
                    },
                },
            },
        }
    )

    engine = create_async_engine(settings.sqlalchemy.connections.default.url)

    sessionmaker = await sessionmaker_service(settings, {"default": engine})
    async with sessionmaker() as session:
        assert session.info == {"framework": "zayt"}


async def test_session_class():
    class MySession(AsyncSession):
        pass

    settings = Settings(
        default_settings
        | {
            "sqlalchemy": {
                "connections": {
                    "default": {
                        "url": "sqlite+aiosqlite:///:memory:",
                    },
                },
                "session": {
                    "options": {
                        "class": MySession,
                    },
                },
            },
        }
    )

    engine = create_async_engine(settings.sqlalchemy.connections.default.url)

    sessionmaker = await sessionmaker_service(settings, {"default": engine})
    async with sessionmaker() as session:
        assert session.__class__ == MySession


@pytest.mark.parametrize(
    "base_a, base_b",
    [
        (
            f"{BaseA.__module__}:{BaseA.__qualname__}",
            f"{BaseB.__module__}:{BaseB.__qualname__}",
        ),
        (BaseA, BaseB),
    ],
    ids=["str", "class"],
)
async def test_binds(base_a, base_b):
    settings = Settings(
        default_settings
        | {
            "sqlalchemy": {
                "connections": {
                    "conn_a": {
                        "url": "sqlite+aiosqlite:///:memory:",
                    },
                    "conn_b": {
                        "url": "sqlite+aiosqlite:///:memory:",
                    },
                },
                "session": {
                    "binds": {
                        base_a: "conn_a",
                        base_b: "conn_b",
                    },
                },
            },
        }
    )

    engine_a = create_async_engine(settings.sqlalchemy.connections.conn_a.url)
    engine_b = create_async_engine(settings.sqlalchemy.connections.conn_b.url)

    sessionmaker = await sessionmaker_service(
        settings, {"conn_a": engine_a, "conn_b": engine_b}
    )
    async with sessionmaker() as session:
        assert session.binds[BaseA] == engine_a
        assert session.binds[BaseB] == engine_b


async def test_sessiomaker_without_default_connection(log_output):
    import structlog

    structlog.configure()

    settings = Settings(
        default_settings
        | {
            "sqlalchemy": {
                "connections": {
                    "conn": {
                        "url": "sqlite+aiosqlite:///:memory:",
                    },
                },
            },
        }
    )

    engine = create_async_engine(settings.sqlalchemy.connections.conn.url)

    sessionmaker = await sessionmaker_service(settings, {"conn": engine})
    async with sessionmaker() as session:
        assert session.bind is engine


@pytest.mark.parametrize(
    "base_a, base_b",
    [
        (
            f"{BaseA.__module__}:{BaseA.__qualname__}",
            f"{BaseB.__module__}:{BaseB.__qualname__}",
        ),
        (BaseA, BaseB),
    ],
    ids=["str", "class"],
)
async def test_binds_model(base_a, base_b):
    settings = Settings(
        default_settings
        | {
            "sqlalchemy": {
                "connections": {
                    "conn_a": {
                        "url": "sqlite+aiosqlite:///:memory:",
                    },
                    "conn_b": {
                        "url": "sqlite+aiosqlite:///:memory:",
                    },
                },
                "session": {
                    "binds": {
                        base_a: "conn_a",
                        base_b: "conn_b",
                    },
                },
            },
        }
    )

    engine_a = create_async_engine(settings.sqlalchemy.connections.conn_a.url)
    async with engine_a.connect() as conn:
        await conn.run_sync(BaseA.metadata.create_all)

    engine_b = create_async_engine(settings.sqlalchemy.connections.conn_b.url)
    async with engine_b.connect() as conn:
        await conn.run_sync(BaseB.metadata.create_all)

    sessionmaker = await sessionmaker_service(
        settings, {"conn_a": engine_a, "conn_b": engine_b}
    )
    async with sessionmaker() as session:
        session.add_all(
            [
                ModelA(id=1, name="A"),
                ModelB(id=1, name="B"),
            ]
        )
        await session.commit()

    async with sessionmaker() as session:
        name_a = await session.scalar(select(ModelA.name).where(ModelA.id == 1))
        name_b = await session.scalar(select(ModelB.name).where(ModelB.id == 1))

        assert name_a == "A"
        assert name_b == "B"


async def test_binds_with_invalid_connection_should_fail():
    settings = Settings(
        default_settings
        | {
            "sqlalchemy": {
                "connections": {
                    "default": {
                        "url": "sqlite+aiosqlite:///:memory:",
                    },
                },
                "session": {
                    "binds": {
                        f"{BaseA.__module__}:{BaseA.__qualname__}": "invalid",
                    },
                },
            },
        }
    )

    with pytest.raises(ValueError, match="No engine with name 'invalid'"):
        await sessionmaker_service(settings, {"default": None})



from collections.abc import Callable
from typing import Any

from zayt.util.maybe_async import call_maybe_async
from zayt.di.container import Container
from zayt.di.service.parse import get_dependencies


async def call_with_dependencies(
    di: Container,
    target: Callable,
    *,
    args: list | tuple = None,
    context: Any = None,
):
    args = args or []

    dependencies = {
        name: await di.get(
            dep.service, name=dep.name, optional=dep.optional, context=context
        )
        for name, dep in get_dependencies(target, skip=len(args))
    }

    return await call_maybe_async(target, *args, **dependencies)



from typing import Self


class Inject:
    """Defines a service dependency"""

    def __init__(self, name: str):
        self.name = name

    def __class_getitem__(cls, item: str) -> Self:
        if not isinstance(item, str):
            raise TypeError()

        return Inject(name=item)



"""
The processing of settings occur as follows:

1. Get main settings values
2. Recursively merge values with settings from profiles, in the order of declaration
"""

import copy
import importlib.util
import os
from collections.abc import Mapping
from copy import deepcopy
from pathlib import Path
from typing import Any

from zayt.conf.defaults import default_settings

__all__ = (
    "Settings",
    "SettingsError",
    "get_settings",
    "get_settings_meta",
    "get_settings_for_profile",
)

SETTINGS_DIR_ENV = "ZAYT_SETTINGS_DIR"
SETTINGS_FILE_ENV = "ZAYT_SETTINGS_FILE"

DEFAULT_SETTINGS_DIR = str(Path("configuration"))
DEFAULT_SETTINGS_FILE = "settings.py"
SETTINGS_KEY = "settings"

ZAYT_PROFILE_ENV = "ZAYT_PROFILE"


class Settings(Mapping[str, Any]):
    def __init__(self, data: dict):
        self._original_data = copy.deepcopy(data)
        self.__data = data

        for key, value in data.items():
            if isinstance(value, dict):
                self.__data[key] = Settings(value)

    def __getattr__(self, item: str):
        try:
            return self.__data[item]
        except KeyError:
            # pylint: disable=raise-missing-from
            raise AttributeError(item)

    def __len__(self) -> int:
        return len(self.__data)

    def __iter__(self):
        return iter(self.__data)

    def __contains__(self, key: str):
        return key in self.__data

    def __getitem__(self, key: str):
        return self.__data[key]

    def __copy__(self):
        return Settings(copy.copy(self._original_data))

    def __deepcopy__(self, memodict):
        data = copy.deepcopy(self._original_data, memodict)
        return Settings(data)

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, Settings):
            return self._original_data == other._original_data
        if isinstance(other, Mapping):
            return self._original_data == other

        return False

    def __str__(self):
        return str(self.__data)

    def __repr__(self):
        return repr(self.__data)


class SettingsError(Exception):
    def __init__(self, path: Path):
        super().__init__(f"cannot load settings from {path}")
        self.path = path


def get_settings() -> Settings:
    """Get settings from files and environment"""
    return get_settings_meta()[0]


def get_settings_meta() -> tuple[Settings, list[tuple[Path, str | None, bool]]]:
    """Get settings from files and environment with metadata

    Metadata contains the files used to load settings from and if the file exists or not
    """
    # get default settings
    settings = deepcopy(default_settings)
    settings_files = []

    # merge with main settings file (settings.yaml)
    profile_settings, profile_settings_file = get_settings_for_profile()
    _merge_recursive(settings, profile_settings)
    settings_files.append(profile_settings_file)

    # merge with profile settings files (settings_$ZAYT_PROFILE.yaml)
    if active_profile_list := os.getenv(ZAYT_PROFILE_ENV):
        for active_profile in active_profile_list.split(","):
            active_profile = active_profile.strip()
            profile_settings, profile_settings_file = get_settings_for_profile(
                active_profile
            )
            _merge_recursive(settings, profile_settings)
            settings_files.append(profile_settings_file)

    return Settings(settings), settings_files


def get_settings_for_profile(
    profile: str = None,
) -> tuple[dict, tuple[Path, str | None, bool]]:
    settings_file = os.getenv(SETTINGS_FILE_ENV, DEFAULT_SETTINGS_FILE)
    settings_dir_path = Path(os.getenv(SETTINGS_DIR_ENV, DEFAULT_SETTINGS_DIR))
    settings_file_path = settings_dir_path / settings_file

    if profile:
        settings_file_path = settings_file_path.with_stem(
            f"{settings_file_path.stem}_{profile}"
        )

    settings_file_path = settings_file_path.absolute()

    try:
        spec = importlib.util.spec_from_file_location(
            "settings" if not profile else f"settings_{profile}",
            settings_file_path,
        )
        settings_module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(settings_module)

        result = getattr(settings_module, SETTINGS_KEY, {})
        if not isinstance(result, dict):
            raise TypeError("invalid settings object type")
        return result, (settings_file_path, profile, True)
    except FileNotFoundError:
        return {}, (settings_file_path, profile, False)
    except TypeError as err:
        raise SettingsError(settings_file_path) from err
    except Exception as err:
        raise SettingsError(settings_file_path) from err


def _merge_recursive(destination: dict, source: dict):
    for key in source:
        if key in destination and all(
            isinstance(arg[key], dict) for arg in (destination, source)
        ):
            _merge_recursive(destination[key], source[key])
        else:
            destination[key] = deepcopy(source[key])



from typing import Annotated

import pytest

from zayt.di.error import (
    FactoryMissingReturnTypeError,
    InvalidDependencyAnnotationError,
    ServiceWithUntypedDependencyError,
)
from zayt.di.inject import Inject
from zayt.di.service.model import ServiceDependency, ServiceSpec
from zayt.di.service.parse import parse_service_spec


def test_parse_service_spec():
    def factory(dependency: Annotated[str, Inject]) -> object:
        pass

    result = parse_service_spec(factory)
    assert result == ServiceSpec(
        service=object,
        impl=None,
        factory=factory,
        name=None,
        scoped=False,
        dependencies=[("dependency", ServiceDependency(service=str, name=None))],
        initializer=None,
        finalizer=None,
    )


def test_factory_missing_return_type_should_fail():
    def factory():
        pass

    with pytest.raises(FactoryMissingReturnTypeError):
        parse_service_spec(factory)


def test_parse_service_spec_dependency_without_annotation():
    def factory(dependency: str) -> object:
        pass

    result = parse_service_spec(factory)
    assert result == ServiceSpec(
        service=object,
        impl=None,
        factory=factory,
        name=None,
        scoped=False,
        dependencies=[("dependency", ServiceDependency(service=str, name=None))],
        initializer=None,
        finalizer=None,
    )


def test_parse_service_spec_with_name():
    def factory() -> object:
        pass

    result = parse_service_spec(factory, name="service")
    assert result == ServiceSpec(
        service=object,
        impl=None,
        factory=factory,
        name="service",
        scoped=False,
        dependencies=[],
        initializer=None,
        finalizer=None,
    )


def test_parse_service_spec_more_dependencies():
    def factory(dependency: Annotated[str, Inject], other: int) -> object:
        pass

    result = parse_service_spec(factory)
    assert result == ServiceSpec(
        service=object,
        impl=None,
        factory=factory,
        name=None,
        scoped=False,
        dependencies=[
            ("dependency", ServiceDependency(service=str, name=None)),
            ("other", ServiceDependency(service=int, name=None)),
        ],
        initializer=None,
        finalizer=None,
    )


def test_parse_service_spec_optional_dependency():
    def factory(dependency: Annotated[str, Inject] = None) -> object:
        pass

    result = parse_service_spec(factory)
    assert result == ServiceSpec(
        service=object,
        impl=None,
        factory=factory,
        name=None,
        scoped=False,
        dependencies=[
            ("dependency", ServiceDependency(service=str, name=None, optional=True))
        ],
        initializer=None,
        finalizer=None,
    )


def test_parse_service_spec_named_dependency():
    def factory(dependency: Annotated[str, Inject("test")]) -> object:
        pass

    result = parse_service_spec(factory)
    assert result == ServiceSpec(
        service=object,
        impl=None,
        factory=factory,
        name=None,
        scoped=False,
        dependencies=[("dependency", ServiceDependency(service=str, name="test"))],
        initializer=None,
        finalizer=None,
    )


def test_parse_service_with_annotated_non_dependency_should_fail():
    def factory(
        _dependency: Annotated[str, Inject], _aproperty: Annotated[int, object]
    ) -> object:
        pass

    with pytest.raises(InvalidDependencyAnnotationError):
        parse_service_spec(factory)


def test_factory_with_untyped_dependency_should_fail():
    def factory(_dependency) -> object:
        pass

    with pytest.raises(ServiceWithUntypedDependencyError):
        parse_service_spec(factory)



from sqlalchemy import text
from sqlalchemy.ext.asyncio import AsyncSession

from zayt.web import startup


@startup
async def on_startup(session: AsyncSession):
    await session.scalar(text("select 1"))



from typing import Annotated

import pytest

from zayt.di.container import Container
from zayt.di.error import (
    IncompatibleDependencyScopeError,
    ScopedServiceWithoutContextError,
)
from zayt.di.inject import Inject


async def test_service_with_different_scope(ioc: Container):
    class ApplicationScopedService:
        pass

    class ContextScopedService:
        dependency: Annotated[ApplicationScopedService, Inject]

    ioc.register(ApplicationScopedService)
    ioc.register(ContextScopedService, scoped=True)

    context = object()

    result = await ioc.get(ContextScopedService, context=context)
    assert isinstance(result.dependency, ApplicationScopedService)


async def test_get_context_scoped_service_without_context_should_fail(ioc: Container):
    class ContextScopedService:
        pass

    ioc.register(ContextScopedService, scoped=True)

    with pytest.raises(ScopedServiceWithoutContextError):
        await ioc.get(ContextScopedService)


async def test_service_with_incompatible_scoped_dependency_should_fail(
    ioc: Container,
):
    class ContextScopedService:
        pass

    class ApplicationScopedService:
        dependency: Annotated[ContextScopedService, Inject]

    ioc.register(ContextScopedService, scoped=True)
    ioc.register(ApplicationScopedService)

    with pytest.raises(IncompatibleDependencyScopeError):
        await ioc.get(ApplicationScopedService)



import os
from importlib.util import find_spec

import pytest
from sqlalchemy import make_url, text
from sqlalchemy.ext.asyncio import create_async_engine

from zayt.conf.defaults import default_settings
from zayt.conf.settings import Settings
from zayt.ext.sqlalchemy.service import make_engine_service, sessionmaker_service

from .test_service import _test_engine_service

POSTGRES_URL = os.getenv("POSTGRES_URL")

pytestmark = [
    pytest.mark.skipif(POSTGRES_URL is None, reason="POSTGRES_URL not defined"),
    pytest.mark.skipif(find_spec("asyncpg") is None, reason="asyncpg not present"),
]

SA_DB_URL = make_url(POSTGRES_URL) if POSTGRES_URL else None


async def test_make_engine_service_with_url():
    settings = Settings(
        default_settings
        | {
            "sqlalchemy": {
                "connections": {
                    "default": {
                        "url": POSTGRES_URL,
                    },
                },
            },
        }
    )

    await _test_engine_service(settings)


async def test_make_engine_service_with_url_username_password():
    settings = Settings(
        default_settings
        | {
            "sqlalchemy": {
                "connections": {
                    "default": {
                        "url": (
                            "postgresql+asyncpg://"
                            f"{SA_DB_URL.host}:{SA_DB_URL.port}/{SA_DB_URL.database}"
                        ),
                        "username": SA_DB_URL.username,
                        "password": SA_DB_URL.password,
                    },
                },
            },
        }
    )

    await _test_engine_service(settings)


async def test_make_engine_service_with_url_components():
    settings = Settings(
        default_settings
        | {
            "sqlalchemy": {
                "connections": {
                    "default": {
                        "drivername": "postgresql+asyncpg",
                        "host": SA_DB_URL.host,
                        "port": SA_DB_URL.port,
                        "database": SA_DB_URL.database,
                        "username": SA_DB_URL.username,
                        "password": SA_DB_URL.password,
                    },
                },
            },
        }
    )

    await _test_engine_service(settings)


async def test_make_engine_service_with_options():
    settings = Settings(
        default_settings
        | {
            "sqlalchemy": {
                "connections": {
                    "default": {
                        "url": POSTGRES_URL,
                        "options": {
                            "echo": True,
                            "echo_pool": True,
                        },
                    },
                },
            },
        }
    )

    engine_service = make_engine_service("default")(settings)
    async for engine in engine_service:
        assert engine.echo is True
        assert engine.pool.echo is True


async def test_make_engine_service_with_execution_options():
    settings = Settings(
        default_settings
        | {
            "sqlalchemy": {
                "connections": {
                    "default": {
                        "url": POSTGRES_URL,
                        "options": {
                            "execution_options": {"isolation_level": "READ COMMITTED"},
                        },
                    },
                },
            },
        }
    )

    async for engine in make_engine_service("default")(settings):
        async with engine.connect() as conn:
            result = await conn.execute(text("select 1"))
            isolation_level = result.context.execution_options["isolation_level"]
            assert isolation_level == "READ COMMITTED"


async def test_sessionmaker_service():
    settings = Settings(
        default_settings
        | {
            "sqlalchemy": {
                "connections": {
                    "default": {
                        "url": POSTGRES_URL,
                    },
                },
            },
        }
    )

    engine = create_async_engine(POSTGRES_URL)

    sessionmaker = await sessionmaker_service(settings, {"default": engine})
    async with sessionmaker() as session:
        result = await session.execute(text("select 1"))
        assert result.scalar() == 1

    await engine.dispose()



import asyncio
from importlib.util import find_spec
import multiprocessing as mp
import os
from pathlib import Path

from httpx import ASGITransport, AsyncClient
import pytest

from .util import with_app


@with_app
async def _run(app):
    async with AsyncClient(
        transport=ASGITransport(app=app), base_url="http://server"
    ) as client:
        for i in range(1, 4):
            response = await client.get("/")
            assert response.json() == {"number": i}


def run():
    return asyncio.run(_run())


MEMCACHED_HOST = os.environ.get("MEMCACHED_HOST")


@pytest.mark.skipif(MEMCACHED_HOST is None, reason="MEMCACHED_HOST not defined")
@pytest.mark.skipif(find_spec("aiomcache") is None, reason="aiomcache not present")
def test_example(monkeypatch):
    path = Path(__file__).parents[2] / "examples" / "memcached"
    monkeypatch.syspath_prepend(path)
    monkeypatch.chdir(path)

    with mp.Pool(processes=1) as pool:
        pool.apply(run)



import asyncio
import multiprocessing as mp
from pathlib import Path

from httpx import AsyncClient
from httpx_ws import aconnect_ws
from httpx_ws.transport import ASGIWebSocketTransport

from .util import with_app


@with_app
async def _run(app):
    async with AsyncClient(
        transport=ASGIWebSocketTransport(app), base_url="http://server"
    ) as client:
        response = await client.get("")
        assert response.status_code == 200
        assert 'new WebSocket("ws://localhost:8000/chat");' in response.text

        async with aconnect_ws("/chat", client) as ws:
            await ws.send_text("Hello World!")
            message = await ws.receive_text()
            assert message == "Hello World!"


def run():
    return asyncio.run(_run())


def test_example(monkeypatch):
    path = Path(__file__).parents[2] / "examples" / "websocket"
    monkeypatch.syspath_prepend(path)
    monkeypatch.chdir(path)

    with mp.Pool(processes=1) as pool:
        pool.apply(run)



import os
from importlib.util import find_spec

import pytest

from zayt.conf import Settings
from zayt.conf.defaults import default_settings
from zayt.ext.memcached.service import make_service

MEMCACHED_HOST = os.getenv("MEMCACHED_HOST")

pytestmark = [
    pytest.mark.skipif(MEMCACHED_HOST is None, reason="MEMCACHED_HOST not defined"),
    pytest.mark.skipif(find_spec("aiomcache") is None, reason="aiomcache not present"),
]


async def _test_make_service(settings: Settings):
    service = make_service("default")(settings)
    async for memcached in service:
        await memcached.set(b"test", b"test")
        result = await memcached.get(b"test")
        assert result == b"test"
        await memcached.delete(b"test")


async def test_make_service_with_address():
    settings = Settings(
        default_settings
        | {
            "memcached": {
                "default": {
                    "host": MEMCACHED_HOST,
                },
            },
        }
    )

    await _test_make_service(settings)


async def test_make_service_with_options():
    settings = Settings(
        default_settings
        | {
            "memcached": {
                "default": {
                    "host": MEMCACHED_HOST,
                    "pool_size": 10,
                    "pool_minsize": 1,
                },
            },
        }
    )

    service = make_service("default")(settings)
    async for memcached in service:
        # pylint: disable=protected-access
        assert memcached._pool._maxsize == 10
        # pylint: disable=protected-access
        assert memcached._pool._minsize == 1



import pytest
import structlog
from structlog.testing import LogCapture


@pytest.fixture(name="log_output")
def fixture_log_output():
    return LogCapture()


@pytest.fixture(autouse=True)
def fixture_configure_structlog(log_output):
    structlog.configure(processors=[log_output])



import inspect
from types import ModuleType

from zayt.di.container import Container
from zayt.di.decorator import ATTRIBUTE_DI_SERVICE
from zayt.util.package_scan import scan_packages


def _is_service(arg) -> bool:
    return (inspect.isfunction(arg) or inspect.isclass(arg)) and hasattr(
        arg, ATTRIBUTE_DI_SERVICE
    )


def discover_services(container: Container, *modules: str | ModuleType):
    for item in scan_packages(*modules, predicate=_is_service):
        service_info = getattr(item, ATTRIBUTE_DI_SERVICE, None)

        container.register(
            item,
            provides=service_info.provides,
            name=service_info.name,
            scoped=service_info.scoped,
        )



import copy

from httpx import ASGITransport, AsyncClient

from zayt.conf import Settings
from zayt.conf.defaults import default_settings
from zayt.web.application import Zayt

from .application import DerivedException, MyBaseException, MyException

SETTINGS = Settings(
    default_settings
    | {
        "modules": [f"{__package__}.application"],
    }
)


async def test_exception_handler():
    settings = copy.copy(SETTINGS)

    app = Zayt(settings)
    # pylint: disable=protected-access
    await app._lifespan_startup()

    client = AsyncClient(transport=ASGITransport(app=app))
    response = await client.get("http://localhost:8000/")
    assert response.json() == {"exception": MyException.__name__}


async def test_derived_exception_handler():
    settings = copy.copy(SETTINGS)

    app = Zayt(settings)
    # pylint: disable=protected-access
    await app._lifespan_startup()

    client = AsyncClient(transport=ASGITransport(app=app))

    response = await client.get("http://localhost:8000/base")
    assert response.text == f"handler=base; exception={MyBaseException.__name__}"

    response = await client.get("http://localhost:8000/derived")
    assert response.text == f"handler=base; exception={DerivedException.__name__}"



from sqlalchemy import text
from sqlalchemy.ext.asyncio import AsyncEngine, create_async_engine

from zayt.conf.defaults import default_settings
from zayt.conf.settings import Settings
from zayt.di.container import Container
from zayt.ext.sqlalchemy.service import (
    engine_dict_service,
    make_engine_service,
    sessionmaker_service,
)


async def _test_engine_service(settings: Settings):
    engine_service = make_engine_service("default")(settings)
    async for engine in engine_service:
        async with engine.connect() as conn:
            result = await conn.execute(text("select 1"))
            assert result.scalar() == 1

        await engine.dispose()


async def test_make_engine_service_with_url():
    settings = Settings(
        default_settings
        | {
            "sqlalchemy": {
                "connections": {
                    "default": {
                        "url": "sqlite+aiosqlite:///:memory:",
                    },
                },
            },
        }
    )

    await _test_engine_service(settings)


async def test_make_engine_service_with_url_components():
    settings = Settings(
        default_settings
        | {
            "sqlalchemy": {
                "connections": {
                    "default": {
                        "drivername": "sqlite+aiosqlite",
                        "database": ":memory:",
                    },
                },
            },
        }
    )

    await _test_engine_service(settings)


async def test_make_engine_service_with_options():
    settings = Settings(
        default_settings
        | {
            "sqlalchemy": {
                "connections": {
                    "default": {
                        "url": "sqlite+aiosqlite:///:memory:",
                        "options": {
                            "echo": True,
                            "echo_pool": True,
                        },
                    },
                },
            },
        }
    )

    engine_service = make_engine_service("default")(settings)
    async for engine in engine_service:
        assert engine.echo is True
        assert engine.pool.echo is True


async def test_make_engine_service_with_execution_options():
    settings = Settings(
        default_settings
        | {
            "sqlalchemy": {
                "connections": {
                    "default": {
                        "url": "sqlite+aiosqlite:///:memory:",
                        "options": {
                            "execution_options": {
                                "isolation_level": "READ UNCOMMITTED"
                            },
                        },
                    },
                },
            },
        }
    )

    async for engine in make_engine_service("default")(settings):
        async with engine.connect() as conn:
            result = await conn.execute(text("select 1"))
            isolation_level = result.context.execution_options["isolation_level"]
            assert isolation_level == "READ UNCOMMITTED"

        await engine.dispose()


async def test_make_engine_service_alternative_name():
    settings = Settings(
        default_settings
        | {
            "sqlalchemy": {
                "connections": {
                    "other": {
                        "url": "sqlite+aiosqlite:///:memory:",
                    },
                },
            },
        }
    )

    engine_service = make_engine_service("other")(settings)
    async for engine in engine_service:
        assert engine is not None
        await engine.dispose()


async def test_sessionmaker_service():
    settings = Settings(
        default_settings
        | {
            "sqlalchemy": {
                "connections": {
                    "default": {
                        "url": "sqlite+aiosqlite:///:memory:",
                    },
                },
            },
        }
    )

    engine = create_async_engine(settings.sqlalchemy.connections.default.url)

    sessionmaker = await sessionmaker_service(settings, {"default": engine})
    async with sessionmaker() as session:
        result = await session.execute(text("select 1"))
        assert result.scalar() == 1

    await engine.dispose()


async def test_engine_dict_service():
    ioc = Container()

    settings = Settings(
        default_settings
        | {
            "sqlalchemy": {
                "connections": {
                    "": {
                        "url": "sqlite+aiosqlite:///:memory:",
                    },
                    "other": {
                        "url": "sqlite+aiosqlite:///:memory:",
                    },
                },
            },
        }
    )

    ioc.define(Settings, settings)
    ioc.define(
        AsyncEngine,
        create_async_engine(settings.sqlalchemy.connections[""].url),
    )
    ioc.define(
        AsyncEngine,
        create_async_engine(settings.sqlalchemy.connections.other.url),
        name="other",
    )

    engines = await engine_dict_service(settings, ioc)

    assert set(engines.keys()) == {"", "other"}



from asgikit import Request

from zayt.ext.jinja import JinjaTemplate
from zayt.web import get


@get("/render")
async def render(request: Request, template: JinjaTemplate):
    await template.respond(request, "template.html", {"variable": "Jinja"})


@get("/stream")
async def stream(request: Request, template: JinjaTemplate):
    await template.respond(
        request,
        "template.html",
        {"variable": "Jinja"},
        stream=True,
    )


@get("/content_type")
async def content_type(request: Request, template: JinjaTemplate):
    await template.respond(
        request,
        "template.html",
        {"variable": "Jinja"},
        media_type="text/plain",
    )



import copy
from http import HTTPStatus
from pathlib import Path

import pytest
from httpx import ASGITransport, AsyncClient

from zayt.conf import Settings
from zayt.conf.defaults import default_settings
from zayt.web.application import Zayt
from zayt.web.middleware.files import static_files_middleware

MIDDLEWARE = [
    f"{static_files_middleware.__module__}:{static_files_middleware.__name__}"
]


@pytest.fixture(autouse=True)
def chdir_fixture(monkeypatch):
    monkeypatch.chdir(Path(__file__).parent)


async def test_static_file():
    settings = Settings(
        default_settings
        | {
            "modules": [f"{__package__}.application"],
            "asgi_middleware": copy.copy(MIDDLEWARE),
        }
    )
    app = Zayt(settings)
    # pylint: disable=protected-access
    await app._lifespan_startup()

    client = AsyncClient(transport=ASGITransport(app=app))
    response = await client.get("http://localhost:8000/static/lorem-ipsum.txt")

    assert response.status_code == HTTPStatus.OK
    assert "text/plain" in response.headers["content-type"]
    assert response.text == "Lorem ipsum dolor sit amet."


async def test_static_file_mapping():
    settings = Settings(
        default_settings
        | {
            "modules": [f"{__package__}.application"],
            "asgi_middleware": copy.copy(MIDDLEWARE),
            "staticfiles": default_settings["staticfiles"]
            | {
                "mappings": {
                    "text-file": "lorem-ipsum.txt",
                },
            },
        }
    )
    app = Zayt(settings)
    # pylint: disable=protected-access
    await app._lifespan_startup()

    client = AsyncClient(transport=ASGITransport(app=app))
    response = await client.get("http://localhost:8000/text-file")

    assert response.request.url.path == "/text-file"
    assert response.status_code == HTTPStatus.OK
    assert "text/plain" in response.headers["Content-Type"]
    assert response.text == "Lorem ipsum dolor sit amet."


async def test_static_files_path():
    settings = Settings(
        default_settings
        | {
            "modules": [f"{__package__}.application"],
            "asgi_middleware": copy.copy(MIDDLEWARE),
            "staticfiles": default_settings["staticfiles"]
            | {
                "path": "assets",
            },
        }
    )
    app = Zayt(settings)
    # pylint: disable=protected-access
    await app._lifespan_startup()

    client = AsyncClient(transport=ASGITransport(app=app))
    response = await client.get("http://localhost:8000/static/lorem-ipsum.txt")
    assert response.status_code == HTTPStatus.NOT_FOUND

    response = await client.get("http://localhost:8000/assets/lorem-ipsum.txt")
    assert response.status_code == HTTPStatus.OK
    assert "text/plain" in response.headers["Content-Type"]
    assert response.text == "Lorem ipsum dolor sit amet."


async def test_static_files_root():
    settings = Settings(
        default_settings
        | {
            "modules": [f"{__package__}.application"],
            "asgi_middleware": copy.copy(MIDDLEWARE),
            "staticfiles": default_settings["staticfiles"]
            | {
                "root": Path(__file__).parent / "resources" / "assets",
            },
        }
    )
    app = Zayt(settings)
    # pylint: disable=protected-access
    await app._lifespan_startup()

    client = AsyncClient(transport=ASGITransport(app=app))
    response = await client.get("http://localhost:8000/static/style.css")
    assert response.status_code == HTTPStatus.OK
    assert "text/css" in response.headers["Content-Type"]
    assert response.text == "body { display: none }"



from typing import Annotated

from asgikit import Request
from sqlalchemy import text
from sqlalchemy.ext.asyncio import AsyncSession

from zayt.di import Inject, service
from zayt.web import get


@service(scoped=True)
class MyService:
    session: Annotated[AsyncSession, Inject]

    async def select(self):
        return await self.session.scalar(text("select 1"))


@get
async def index(request: Request, my_service: MyService):
    value = await my_service.select()
    await request.respond_text(str(value))



from http import HTTPStatus

import pytest

from zayt.web.exception import (
    HTTPBadRequestException,
    HTTPStatusException,
    HTTPForbiddenException,
    HTTPInternalServerException,
    HTTPNotFoundException,
    HTTPUnauthorizedException,
)


@pytest.mark.parametrize(
    "error_cls, status",
    [
        (HTTPBadRequestException, HTTPStatus.BAD_REQUEST),
        (HTTPNotFoundException, HTTPStatus.NOT_FOUND),
        (HTTPUnauthorizedException, HTTPStatus.UNAUTHORIZED),
        (HTTPForbiddenException, HTTPStatus.FORBIDDEN),
        (HTTPInternalServerException, HTTPStatus.INTERNAL_SERVER_ERROR),
    ],
    ids=[
        HTTPStatus.BAD_REQUEST.name,
        HTTPStatus.NOT_FOUND.name,
        HTTPStatus.UNAUTHORIZED.name,
        HTTPStatus.FORBIDDEN.name,
        HTTPStatus.INTERNAL_SERVER_ERROR.name,
    ],
)
def test_correct_status(error_cls, status: HTTPStatus):
    error = error_cls()
    assert error.status == status


def test_custom_http_exception():
    error = HTTPStatusException(status=HTTPStatus.IM_A_TEAPOT)
    assert error.status == HTTPStatus.IM_A_TEAPOT



import inspect

from zayt.util.package_scan import scan_packages
from zayt.web.routing.decorator import (
    ATTRIBUTE_HANDLER,
    ATTRIBUTE_WEBSOCKET,
    HandlerInfo,
    WebSocketInfo,
)


def _is_handler(arg) -> bool:
    return inspect.iscoroutinefunction(arg) and (
        hasattr(arg, ATTRIBUTE_HANDLER) or hasattr(arg, ATTRIBUTE_WEBSOCKET)
    )


def discover_handlers(self, *modules):
    for item in scan_packages(*modules, predicate=_is_handler):
        handler, prefix = item if isinstance(item, tuple) else (item, None)

        handler_info: HandlerInfo = getattr(handler, ATTRIBUTE_HANDLER, None)
        websocket_info: WebSocketInfo = getattr(handler, ATTRIBUTE_WEBSOCKET, None)

        if handler_info:
            for method, path in handler_info.mappings:
                self.route(handler, method, path, prefix=prefix)

        if websocket_info:
            for path in websocket_info.paths:
                self.websocket(handler, path, prefix=prefix)



from asgikit import Request


class DuplicateRouteError(Exception):
    def __init__(self, route1: str, route2: str):
        super().__init__(f"path for {route1} clashes with {route2}")


class HandlerWithoutDecoratorError(Exception):
    def __init__(self, handler):
        super().__init__(
            f"{handler.__module__}.{handler.__qualname__}"
            " is not decorated with @handler"
        )


class HandlerNotAsyncError(Exception):
    def __init__(self, handler):
        super().__init__(f"{handler.__module__}.{handler.__qualname__} must be async")


class HandlerMissingRequiredArgumentError(Exception):
    def __init__(self, handler):
        super().__init__(
            f"{handler.__module__}.{handler.__qualname__}"
            " must receive the 'request' parameter"
        )


class HandlerRequestTypeError(Exception):
    def __init__(self, handler):
        super().__init__(
            f"{handler.__module__}.{handler.__qualname__}"
            " first parameter annotation must be "
            f"'{Request.__module__}.{Request.__name__}'"
        )


class HandlerUntypedParametersError(Exception):
    def __init__(self, handler, params: list[str]):
        param_list = ", ".join(params)
        super().__init__(
            f"{handler.__module__}.{handler.__qualname__}"
            " parameters must be typed:"
            f" ({param_list})"
        )



import pytest
from httpx import ASGITransport, AsyncClient

from zayt.conf.defaults import default_settings
from zayt.conf.settings import Settings
from zayt.di.error import ScopedServiceWithoutContextError
from zayt.web.application import Zayt


async def test_application():
    settings = Settings(
        default_settings
        | {
            "modules": [
                f"{__package__}.application_session",
                "zayt.ext.sqlalchemy",
            ],
            "sqlalchemy": {
                "connections": {"": {"url": "sqlite+aiosqlite:///:memory:"}}
            },
        }
    )

    app = Zayt(settings)

    # pylint: disable=protected-access
    await app._lifespan_startup()

    client = AsyncClient(transport=ASGITransport(app=app))
    response = await client.get("http://localhost:8000/")

    assert response.text == "1"


async def test_session_outside_request_should_fail():
    settings = Settings(
        default_settings
        | {
            "modules": [
                f"{__package__}.application_session_startup",
                "zayt.ext.sqlalchemy",
            ],
            "sqlalchemy": {
                "connections": {"default": {"url": "sqlite+aiosqlite:///:memory:"}}
            },
        }
    )

    app = Zayt(settings)

    with pytest.raises(ScopedServiceWithoutContextError):
        # pylint: disable=protected-access
        await app._lifespan_startup()



import inspect
import warnings
from collections.abc import Callable, Iterable
from http import HTTPMethod
from typing import NamedTuple

from asgikit import Request

from zayt.web.handler.parse import assert_params_annotated
from zayt.web.routing.exception import (
    HandlerMissingRequiredArgumentError,
    HandlerNotAsyncError,
    HandlerRequestTypeError,
)

ATTRIBUTE_HANDLER = "__zayt_web_action__"
ATTRIBUTE_WEBSOCKET = "__zayt_web_websocket__"


class HandlerInfo(NamedTuple):
    mappings: set[tuple[HTTPMethod, str]]


class WebSocketInfo(NamedTuple):
    paths: set[str]


def _check_handler(handler: Callable):
    if not inspect.iscoroutinefunction(handler):
        raise HandlerNotAsyncError(handler)

    assert_params_annotated(handler, skip=2)

    params = list(inspect.signature(handler).parameters.values())
    if len(params) < 1:
        raise HandlerMissingRequiredArgumentError(handler)

    req_param = params[0].annotation
    if req_param is not inspect.Signature.empty:
        if req_param is not Request:
            raise HandlerRequestTypeError(handler)


def route(method: HTTPMethod | Iterable[HTTPMethod], path: str | None):
    path = path.strip("/") if path else ""

    if isinstance(method, HTTPMethod):
        method = {method}
    elif isinstance(method, (tuple, list)):
        method = set(method)

    def wrapper(handler: Callable):
        _check_handler(handler)

        handler_info = getattr(handler, ATTRIBUTE_HANDLER, None)
        if not handler_info:
            handler_info = HandlerInfo(set())

        for m in method:
            if (m, path) in handler_info.mappings:
                warnings.warn(
                    f"Duplicate annotation in handler {handler.__module__}.{handler.__qualname__}: "
                    f"{m} {path}"
                )

            handler_info.mappings.add((m, path))

        setattr(handler, ATTRIBUTE_HANDLER, handler_info)
        return handler

    return wrapper


def _route(method: HTTPMethod | None, path_or_action: str | Callable):
    if isinstance(path_or_action, str):
        path = path_or_action.strip("/")
        action = None
    else:
        path = ""
        action = path_or_action

    wrapper = route(method=[method], path=path)
    return wrapper(action) if action else wrapper


def get(path_or_action: str | Callable):
    return _route(HTTPMethod.GET, path_or_action)


def post(path_or_action: str | Callable):
    return _route(HTTPMethod.POST, path_or_action)


def put(path_or_action: str | Callable):
    return _route(HTTPMethod.PUT, path_or_action)


def patch(path_or_action: str | Callable):
    return _route(HTTPMethod.PATCH, path_or_action)


def delete(path_or_action: str | Callable):
    return _route(HTTPMethod.DELETE, path_or_action)


def websocket(path_or_action: str | Callable):
    if isinstance(path_or_action, str):
        path = path_or_action.strip("/")
        action = None
    else:
        path = ""
        action = path_or_action

    def wrapper(handler: Callable):
        _check_handler(handler)

        websocket_info = getattr(handler, ATTRIBUTE_WEBSOCKET, None)
        if not websocket_info:
            websocket_info = WebSocketInfo(set())

        websocket_info.paths.add(path)
        setattr(handler, ATTRIBUTE_WEBSOCKET, websocket_info)

        return handler

    return wrapper(action) if action else wrapper



from importlib.util import find_spec

from zayt.web.lifecycle.decorator import startup

from .service import JinjaTemplate

__all__ = ("JinjaTemplate",)


@startup
async def init():
    if find_spec("jinja2") is None:
        raise ModuleNotFoundError(
            "Missing 'jinja2'. Install 'zayt' with 'jinja' extra."
        )



from asgikit import Request
from redis.asyncio import Redis

from zayt.web import get


@get
async def index(request: Request, redis: Redis):
    await redis.set("key", "value")
    result = (await redis.get("key")).decode("utf-8")

    await request.respond_text(result)
    await redis.delete("key")



import asyncio
from importlib import import_module
import traceback
from http import HTTPStatus

import structlog
from asgikit import (
    Request,
    Response,
    WebSocket,
    WebSocketState,
    WebSocketDisconnect,
    WebSocketResponseNotSupportedError,
)

from zayt.conf.settings import Settings, get_settings_meta
from zayt.di.call import call_with_dependencies
from zayt.di.container import Container
from zayt.di.discover import discover_services
from zayt.util.base_types import get_base_types
from zayt.util.import_item import import_item
from zayt.web.exception import (
    HTTPStatusException,
    WebSocketException,
    HTTPNotFoundException,
)
from zayt.web.exception_handler.decorator import ExceptionHandlerType
from zayt.web.exception_handler.discover import find_exception_handlers
from zayt.web.lifecycle.discover import find_background_services, find_startup_hooks
from zayt.web.routing.router import Router
from zayt.web.routing.discover import discover_handlers

logger = structlog.get_logger(__name__)


def _init_settings(settings: Settings | None) -> Settings:
    settings_files = []
    if not settings:
        settings, settings_files = get_settings_meta()

    logging_setup = import_item(settings.logging.setup)
    logging_setup(settings)

    for settings_file, profile, found in settings_files:
        if found:
            logger.info("settings loaded", file=settings_file.name)
        elif profile:
            logger.info("settings file not found", file=settings_file.name)

    return settings


# pylint: disable=too-many-instance-attributes
class Zayt:
    """Entrypoint class for a Zayt Application

    Will try to automatically import and register a module called "application".
    Other modules and classes can be registered using the "register" method
    """

    def __init__(self, settings: Settings = None):
        self.settings = _init_settings(settings)

        self.di = Container()
        self.di.define(Container, self.di)
        self.di.define(Settings, self.settings)

        self.router = Router()
        self.di.define(Router, self.router)

        self.handler = self._request_handler

        modules = self.settings.modules
        try:
            import_module("application")
            has_application_module = True
        except ImportError:
            has_application_module = False

        if has_application_module:
            if "application" not in modules:
                modules.append("application")
        elif not modules:
            raise RuntimeError("Missing module 'application'")

        self.exception_handlers = find_exception_handlers(*modules)
        self._exception_handler_map = {}

        self.startup = find_startup_hooks(*modules)

        self.background_services = find_background_services(*modules)
        self._background_services: set[asyncio.Task] = set()

        discover_services(self.di, *modules)
        discover_handlers(self.router, *modules)

    async def __call__(self, scope, receive, send):
        match scope["type"]:
            case "http" | "websocket":
                await self._handle_request(scope, receive, send)
            case "lifespan":
                await self._handle_lifespan(scope, receive, send)
            case _:
                raise RuntimeError(f"unknown scope '{scope['type']}'")

    async def _initialize_asgi_middleware(self):
        middleware = self.settings.asgi_middleware
        if not middleware:
            return

        middleware_functions = [import_item(name) for name in middleware]

        for factory in reversed(middleware_functions):
            self.handler = await call_with_dependencies(
                self.di, factory, args=[self.handler]
            )

    async def _lifespan_startup(self):
        await self._initialize_asgi_middleware()

        for hook in self.startup:
            await call_with_dependencies(self.di, hook)

        for hook in self.background_services:
            task = asyncio.create_task(call_with_dependencies(self.di, hook))
            self._background_services.add(task)

            def done_callback(done):
                if err := done.exception():
                    logger.exception(err, exc_info=err)
                self._background_services.discard(done)

            task.add_done_callback(done_callback)

    async def _lifespan_shutdown(self):
        for task in self._background_services:
            if not task.done():
                task.cancel()

        await self.di.finalize()

    async def _handle_lifespan(self, _scope, receive, send):
        while True:
            message = await receive()
            if message["type"] == "lifespan.startup":
                logger.debug("handling lifespan startup")
                try:
                    await self._lifespan_startup()
                    logger.debug("lifespan startup complete")
                    await send({"type": "lifespan.startup.complete"})
                except Exception as err:
                    logger.exception("lifespan startup failed")
                    await send({"type": "lifespan.startup.failed", "message": str(err)})
            elif message["type"] == "lifespan.shutdown":
                logger.debug("handling lifespan shutdown")
                try:
                    await self._lifespan_shutdown()
                    logger.debug("lifespan shutdown complete")
                    await send({"type": "lifespan.shutdown.complete"})
                except Exception as err:
                    logger.debug("lifespan shutdown failed")
                    await send(
                        {"type": "lifespan.shutdown.failed", "message": str(err)}
                    )
                break

    async def _handle_request(self, scope, receive, send):
        try:
            await self.handler(scope, receive, send)
        except WebSocketDisconnect:
            client = scope["client"]
            logger.info(
                "websocket disconnect from client",
                client=f"{client[0]}:{client[1]}",
            )
            await WebSocket(scope, receive, send).close()
        except WebSocketException as err:
            if scope["type"] == "http":
                logger.error("request handler raise WebSocketException")
                response = Response(scope, receive, send)
                if not response.is_started:
                    await response.start(HTTPStatus.INTERNAL_SERVER_ERROR)
                    await response.end()
                elif not response.is_finished:
                    await response.end()
                return

            await WebSocket(scope, receive, send).close(err.code, err.reason)
        except HTTPStatusException as err:
            await self._handle_http_exception(scope, receive, send, err)
        except Exception as err:
            logger.exception("error processing request")

            if scope["type"] == "http":
                trace = "".join(traceback.format_exception(err))
                request = Request(scope, receive, send)
                await request.respond_text(trace, status=500)
            else:
                await WebSocket(scope, receive, send).close(1011)

    async def _request_handler(self, scope, receive, send):
        request = Request(scope, receive, send)
        self.di.define(Request, request, context=request)

        path = request.path
        method = request.method

        if request.is_http:
            logger.debug(
                "handling http request",
                method=method,
                path=path,
                query=request.query,
            )
        else:
            logger.debug(
                "handling websocket",
                path=path,
                query=request.query,
            )

        match = self.router.match(method, path)

        if not match:
            raise HTTPNotFoundException()

        handler = match.route.action
        path_params = match.params
        scope["path_params"] = path_params

        try:
            await call_with_dependencies(
                self.di, handler, args=(request,), context=request
            )

            if request.is_websocket:
                websocket = WebSocket(scope, receive, send)
                if websocket.state != WebSocketState.CLOSED:
                    await websocket.close()
            elif not request.response.is_started:
                await request.respond_empty(HTTPStatus.INTERNAL_SERVER_ERROR)
            elif not request.response.is_finished:
                await request.response.end()
        except Exception as err:
            if request.is_websocket:
                raise

            if handler := self._get_exception_handler(type(err)):
                logger.debug(
                    "Handling exception with exception handler",
                    handler=f"{handler.__module__}.{handler.__qualname__}",
                )

                request = Request(scope, receive, send)
                await call_with_dependencies(
                    self.di, handler, args=(request, err), context=request
                )
            else:
                raise
        finally:
            await self.di.finalize(request)

    def _get_exception_handler(
        self, err: type[BaseException]
    ) -> ExceptionHandlerType | None:
        if handler := self._exception_handler_map.get(err):
            return handler

        for base in get_base_types(err):
            if handler := self.exception_handlers.get(base):
                self._exception_handler_map[err] = handler
                return handler

        return None

    @staticmethod
    async def _handle_http_exception(scope, receive, send, err: HTTPStatusException):
        if cause := err.__cause__:
            logger.error("http exception with cause", exc_info=cause)
            stack_trace = "".join(traceback.format_exception(cause))
        else:
            stack_trace = None

        request = Request(scope, receive, send)

        if request.response.is_finished:
            logger.error("request already finished", exc_info=cause)
            return

        try:
            if stack_trace:
                await request.respond_text(
                    stack_trace, status=err.status, media_type="text/plain"
                )
            else:
                await request.respond_empty(err.status)
        except WebSocketResponseNotSupportedError:
            await WebSocket(scope, receive, send).close(1011 if err.__cause__ else 1006)
            return



import json
import os
from collections.abc import Callable
from typing import TypeVar

T = TypeVar("T")


def env(name: str, cast: Callable[[str], T] = str, default: T = None) -> T:
    """Read an environment variable, optionally applying a cast

    :param name: Name of the environment variable
    :param cast: Type to convert the value
        Should be a callable that accepts a ``str`` and returns the converted value.
        :py:type:`bool` receives a special treatment:
        If the value is empty, return ``None``.
        If the value is ``"1"`` or ``"true"`` (ignoring case), the result is ``True``.
        Any other value is ``False``.
        :py:module:`json` also receives a special treatment, in which the value
        is converted with ``json.loads``.
    :param default: Default value if the environment variable is not set
    """

    value = os.environ.get(name)

    if value is None:
        if default is not None:
            return default
        raise ValueError(f"Environment variable '{name}' not set")

    if len(value.strip()) == 0:
        return None

    if cast is bool:
        return _cast_bool(value)

    if cast is json:
        return json.loads(value)

    return cast(value)


def _cast_bool(value: str) -> bool | None:
    """Cast value to bool from str

    :return:
        True if `value.lower()` is "true" or "1"
        None if `value.strip` is empty
        False otherwise
    """

    return value.lower() in {"true", "1"}



from httpx import ASGITransport, AsyncClient

from zayt.conf import Settings
from zayt.conf.defaults import default_settings
from zayt.di.decorator import service
from zayt.web.application import Zayt


def my_middleware(app):
    async def inner(scope, receive, send):
        async def new_send(event: dict):
            if event["type"] == "http.response.body":
                event["body"] = b"Middleware Ok"
            await send(event)

        await app(scope, receive, new_send)

    return inner


class MyService:
    pass


def my_middleware_with_dep(app, dep: MyService):
    async def inner(scope, receive, send):
        async def new_send(event: dict):
            if event["type"] == "http.response.body":
                event["body"] = f"Middleware Ok ({dep.__class__.__name__})".encode()
            await send(event)

        await app(scope, receive, new_send)

    return inner


async def test_middleware():
    settings = Settings(
        default_settings
        | {
            "modules": [f"{__package__}.application"],
            "asgi_middleware": [f"{__package__}.test_middleware:my_middleware"],
        }
    )
    app = Zayt(settings)
    # pylint: disable=protected-access
    await app._lifespan_startup()

    client = AsyncClient(transport=ASGITransport(app=app))
    response = await client.get("http://localhost:8000/")
    assert response.text == "Middleware Ok"


async def test_middleware_with_dep():
    settings = Settings(
        default_settings
        | {
            "modules": [f"{__package__}.application"],
            "asgi_middleware": [
                f"{__package__}.test_middleware:my_middleware_with_dep"
            ],
        }
    )
    app = Zayt(settings)
    app.di.register(service(MyService))
    # pylint: disable=protected-access
    await app._lifespan_startup()

    client = AsyncClient(transport=ASGITransport(app=app))
    response = await client.get("http://localhost:8000/")
    assert response.text == f"Middleware Ok ({MyService.__name__})"



from collections.abc import Callable
from http import HTTPMethod

import structlog

from zayt.web.routing.exception import DuplicateRouteError
from zayt.web.routing.route import Route, RouteMatch

logger = structlog.get_logger(__name__)


class Router:
    def __init__(self):
        self.routes: list[Route] = []

    def _check_duplicates(self, route):
        for current_route in self.routes:
            if (
                current_route.action != route.action
                and current_route.method == route.method
                and current_route.regex == route.regex
            ):
                raise DuplicateRouteError(route.name, current_route.name)

    def route(
        self, handler: Callable, method: HTTPMethod, path: str, *, prefix: str = None
    ):
        path = path.strip("/")
        if prefix:
            path = prefix.strip("/") + "/" + path

        route_name = f"{method.lower()}.{handler.__module__}.{handler.__qualname__}"
        route = Route(method, path, handler, route_name)
        self._check_duplicates(route)

        self.routes.append(route)
        logger.debug(
            "action registered",
            handler=f"{handler.__module__}.{handler.__qualname__}",
            method=route.method,
            path=route.path,
        )

        self.routes.sort()

    def websocket(self, handler: Callable, path: str, *, prefix: str = None):
        path = path.strip("/")
        if prefix:
            path = prefix.strip("/") + "/" + path

        route_name = f"websocket.{handler.__module__}.{handler.__qualname__}"
        route = Route(None, path, handler, route_name)
        self._check_duplicates(route)

        self.routes.append(route)
        logger.debug(
            "websocket registered",
            handler=f"{handler.__module__}.{handler.__qualname__}",
            path=route.path,
        )

        self.routes.sort()

    def match(self, method: HTTPMethod | None, path: str) -> RouteMatch | None:
        """Match a path against the routes

        :param method: HTTP method or None for websocket
        :param path: path to match against
        """

        for route in self.routes:
            if (match := route.match(method, path)) is not None:
                return RouteMatch(route, method, path, match)

        return None



import asyncio
from importlib.util import find_spec
import multiprocessing as mp
import os
from pathlib import Path

from httpx import ASGITransport, AsyncClient
import pytest

from .util import with_app


@with_app
async def _run(app):
    async with AsyncClient(
        transport=ASGITransport(app=app), base_url="http://server"
    ) as client:
        response = await client.get("/")
        result = response.json().get("model")
        assert result is not None
        assert result["name"] == "MyModel"

        response = await client.get("/other")
        result = response.json().get("model")
        assert result is not None
        assert result["name"] == "OtherModel"


def run():
    return asyncio.run(_run())


POSTGRES_URL = os.getenv("POSTGRES_URL")


@pytest.mark.skipif(POSTGRES_URL is None, reason="POSTGRES_URL not defined")
@pytest.mark.skipif(find_spec("asyncpg") is None, reason="asyncpg not present")
def test_example(monkeypatch):
    path = Path(__file__).parents[2] / "examples" / "sqlalchemy"
    monkeypatch.syspath_prepend(path)
    monkeypatch.chdir(path)

    with mp.Pool(processes=1) as pool:
        pool.apply(run)



from typing import Annotated

from aiomcache import Client
from asgikit import Request

from zayt.di import Inject
from zayt.web import get


@get
async def index(request: Request, memcached: Annotated[Client, Inject(name="other")]):
    await memcached.set(b"key", b"value")
    result = (await memcached.get(b"key")).decode("utf-8")

    await request.respond_text(result)
    await memcached.delete(b"key")



from collections.abc import Iterator


def get_header(header_name: str, scope: dict, default=None) -> str | None:
    """Retrieve the header value from the asgi scope

    :return: The decoded header value or the given default
    :raises ValueError: If the scope does not have the "headers" key
    """

    return next(get_header_all(header_name, scope), default)


def get_header_all(header_name: str, scope: dict) -> Iterator[str]:
    """Iterate over all header values for the given header name in the asgi scope

    :return: And iterator over the decoded header values
    :raises ValueError: If the scope does not have the "headers" key
    """

    if "headers" not in scope:
        raise ValueError("invalid scope")

    for name, value in scope["headers"]:
        if name.decode("ascii").lower() == header_name:
            yield value.decode("ascii")



import inspect
from typing import Annotated

from zayt.di.decorator import ATTRIBUTE_DI_SERVICE, service
from zayt.di.inject import Inject
from zayt.di.service.model import ServiceInfo


def test_decorator():
    @service
    class Service:
        pass

    assert getattr(Service, ATTRIBUTE_DI_SERVICE) == ServiceInfo(
        None, None, scoped=False
    )


def test_class_with_dependency_annotation():
    class Dependency:
        pass

    @service
    class Service:
        dependency: Annotated[Dependency, Inject]

    init = inspect.signature(Service.__init__)
    assert len(init.parameters) == 3

    assert Service().dependency is None
    assert Service(Dependency()).dependency is not None
    assert Service(dependency=Dependency()).dependency is not None


def test_class_with_non_dependency_annotation():
    class NonDependency:
        pass

    @service
    class Service:
        non_dependency: NonDependency

    assert not hasattr(Service(), "non_dependency")
    assert not hasattr(Service(NonDependency()), "non_dependency")
    assert not hasattr(Service(non_dependency=NonDependency()), "non_dependency")


def test_class_with_mixed_dependency_annotation():
    class Dependency:
        pass

    class NonDependency:
        pass

    @service
    class Service:
        dependency: Annotated[Dependency, Inject]
        non_dependency: NonDependency

    assert Service().dependency is None
    assert Service(Dependency()).dependency is not None
    assert Service(dependency=Dependency()).dependency is not None
    assert not hasattr(Service(), "non_dependency")
    assert not hasattr(Service(NonDependency()), "non_dependency")
    assert not hasattr(Service(non_dependency=NonDependency()), "non_dependency")



import pytest

from zayt.di.container import Container
from zayt.di.error import FactoryMissingReturnTypeError, ServiceAlreadyRegisteredError


class Service1:
    pass


async def service1_factory() -> Service1:
    return Service1()


class Service2:
    def __init__(self, service1: Service1):
        self.service1 = service1


async def service2_factory(service1: Service1) -> Service2:
    return Service2(service1)


class Interface:
    pass


class Implementation(Interface):
    pass


async def interface_factory() -> Interface:
    return Implementation()


def test_has_service(ioc: Container):
    ioc.register(service1_factory)
    assert ioc.has(Service1)


async def test_service_with_provided_interface(ioc: Container):
    ioc.register(interface_factory)

    instance = await ioc.get(Interface)
    assert isinstance(instance, Implementation)


async def test_inject_singleton(ioc: Container):
    ioc.register(service1_factory)
    ioc.register(service2_factory)

    instance = await ioc.get(Service2)
    assert isinstance(instance, Service2)
    assert isinstance(instance.service1, Service1)

    other_instance = await ioc.get(Service2)
    assert other_instance is instance
    assert other_instance.service1 is instance.service1


async def test_interface_implementation(ioc: Container):
    ioc.register(interface_factory)

    instance = await ioc.get(Interface)
    assert isinstance(instance, Implementation)


def test_factory_function_without_return_type_should_fail(ioc: Container):
    async def service_factory():
        pass

    with pytest.raises(FactoryMissingReturnTypeError):
        ioc.register(service_factory)


def test_provides_option_should_log_warning(ioc: Container, log_output):
    async def factory() -> Interface:
        pass

    ioc.register(factory, provides=Interface)

    assert (
        log_output.entries[0]["event"]
        == "option 'provides' on a factory function has no effect"
    )
    assert log_output.entries[0]["service"] == "tests.di.test_service_function.factory"


async def test_sync_factory(ioc: Container):
    def sync_factory() -> Service1:
        return Service1()

    ioc.register(sync_factory)

    instance = await ioc.get(Service1)
    assert isinstance(instance, Service1)


async def test_register_a_service_twice_should_fail(ioc: Container):
    async def duplicate_factory() -> Service1:
        pass

    with pytest.raises(ServiceAlreadyRegisteredError):
        ioc.register(service1_factory)
        ioc.register(duplicate_factory)



import inspect
import typing
from collections.abc import Callable
from functools import cache
from typing import Annotated

from zayt.di.inject import Inject
from zayt.web.handler.model import RequestParam
from zayt.web.routing.exception import HandlerUntypedParametersError


def assert_params_annotated(handler: Callable, *, skip: int):
    """Asserts that parameters past 'skip' are annotated.

    :raises HandlerUntypedParametersError: if parameters are not annotated.
    """
    signature = inspect.signature(handler)
    parameters = list(signature.parameters.values())

    if untyped_params := [
        p.name for p in parameters[skip:] if p.annotation is inspect.Signature.empty
    ]:
        raise HandlerUntypedParametersError(handler, untyped_params)


@cache
def parse_handler_params(
    handler: Callable, *, skip: int
) -> list[tuple[str, RequestParam]]:
    """parse handler parameters from function signature

    returns: list of 2-tuples of parameter name and parameter specification
    :raises HandlerUntypedParametersError:
    """

    assert_params_annotated(handler, skip=skip)

    result = []

    signature = inspect.signature(handler, eval_str=True)
    parameters = list(signature.parameters.keys())

    if len(parameters) == 0:
        return result

    if len(parameters) <= skip:
        return result

    for name in parameters[skip:]:  # skip request parameter
        type_hint = signature.parameters[name].annotation
        has_default = signature.parameters[name].default is not inspect.Parameter.empty

        if typing.get_origin(type_hint) is not Annotated:
            result.append((name, RequestParam(type_hint, None, has_default)))
        else:
            # Annotated is garanteed to have at least 2 args
            param_type, param_meta, *_ = typing.get_args(type_hint)
            if not (param_meta is Inject or isinstance(param_meta, Inject)):
                raise TypeError("Inject")
            result.append((name, RequestParam(param_type, param_meta, has_default)))

    return result



from asgikit import Request
from zayt.web import get, websocket


@get
async def index(request: Request):
    await request.respond_text("Ok")


@websocket("exception/new")
async def ws_exception_new(request: Request):
    raise Exception("NEW")


@websocket("exception/accepted")
async def ws_exception_accepted(request: Request):
    _ws = await request.upgrade()
    raise Exception("accepted")



from aiomcache import Client, FlagClient

from zayt.conf.settings import Settings


def make_service(name: str):
    async def memcached_service(settings: Settings) -> Client:
        memcached_settings = dict(settings.memcached[name])
        args = []
        kwargs = memcached_settings

        if host := kwargs.pop("host", None):
            args.append(host)
            if port := kwargs.pop("port", None):
                args.append(int(port))

        client = FlagClient(*args, **kwargs)

        yield client
        await client.close()

    return memcached_service



import re
import typing
from collections import Counter
from http import HTTPMethod
from re import Pattern
from typing import Callable, NamedTuple

__all__ = ("Route", "RouteMatch")

RE_PATH_PARAM_SPEC = re.compile(r"\{(\*?)([a-zA-Z\w]+)}")
RE_MULTI_SLASH = re.compile(r"/{2,}")

PATH_PARAM_PATTERN = {
    "": r"[^/]+?",
    "*": r".*",
}


def build_path_regex_and_params(
    handler: Callable, path: str
) -> tuple[Pattern, dict[str, type]]:
    """parse path and build regex for route matching

    :returns: compiled regex and tuple of mapping param name to type
    """

    type_hints = typing.get_type_hints(handler)
    type_hints.pop("return", None)

    regex = RE_MULTI_SLASH.sub("/", path)
    path_params = RE_PATH_PARAM_SPEC.findall(path)

    # verify that path does not have duplicate parameters
    counter = Counter(path_params)
    repeated = [name for name, occurrencies in counter.items() if occurrencies > 1]
    if repeated:
        raise ValueError(
            f"path parameters defined more than once: {', '.join(repeated)}"
        )

    param_types = {}
    for kind, param in path_params:
        type_hint = type_hints.get(param, str)
        param_types[param] = type_hint

        param_regex = f"(?P<{param}>{PATH_PARAM_PATTERN[kind]})"
        regex = regex.replace("{" + f"{kind}{param}" + "}", param_regex)

    if not regex.startswith("/"):
        regex = "/" + regex

    if regex.endswith("/"):
        regex += "?"
    else:
        regex += "/?"

    regex = f"^{regex}$"
    return re.compile(regex), param_types


class Route:
    def __init__(
        self,
        method: HTTPMethod | None,
        path: str,
        action: Callable,
        name: str,
    ):
        self.method = method
        self.path = path
        self.action = action
        self.name = name
        self.regex, self.path_params = build_path_regex_and_params(action, path)

    def match(self, method: HTTPMethod | None, path: str) -> dict[str, str] | None:
        if method is self.method and (match := self.regex.match(path)):
            return match.groupdict()

        return None

    def reverse(self, **kwargs) -> str:
        path = self.path

        for param, value in kwargs.items():
            if param not in self.path_params:
                raise ValueError()
            path = path.replace(f"{{{param}}}", value)

        return path

    def __lt__(self, other):
        path1 = self.path.split("/")
        path2 = other.path.split("/")

        if len(path1) != len(path2):
            return path1 < path2

        for i1, i2 in zip(path1, path2):
            if i1.startswith("{") and not i2.startswith("{"):
                return False  # path1 last
            if not i1.startswith("{") and i2.startswith("{"):
                return True  # path2 last

        return path1 < path2


class RouteMatch(NamedTuple):
    route: Route
    method: HTTPMethod | None
    path: str
    params: dict[str, str]



import inspect

from zayt.util.package_scan import scan_packages
from zayt.web.exception_handler.decorator import (
    ATTRIBUTE_EXCEPTION_HANDLER,
    ExceptionHandlerType,
)


def _is_exception_handler(arg) -> bool:
    return inspect.iscoroutinefunction(arg) and hasattr(
        arg, ATTRIBUTE_EXCEPTION_HANDLER
    )


def find_exception_handlers(*modules) -> dict[type[Exception], ExceptionHandlerType]:
    result = {}

    for item in scan_packages(*modules, predicate=_is_exception_handler):
        exc_handler_info = getattr(item, ATTRIBUTE_EXCEPTION_HANDLER)
        exc_type = exc_handler_info.exception_class
        if exc_type in result:
            raise ValueError("Exception handler already registered")
        result[exc_type] = item

    return result



from importlib.util import find_spec

from zayt.conf.settings import Settings
from zayt.di.container import Container
from zayt.web.lifecycle.decorator import startup

from .service import make_engine_service


@startup
def init(container: Container, settings: Settings):
    if find_spec("sqlalchemy") is None:
        raise ModuleNotFoundError(
            "Missing 'sqlalchemy'. Install 'zayt' with 'sqlalchemy' extra."
        )

    for name in settings.sqlalchemy.connections:
        assert isinstance(name, str)
        name = name.strip()
        container.register(make_engine_service(name), name=name if name else None)



from pathlib import Path

from httpx import ASGITransport, AsyncClient

from zayt.conf.defaults import default_settings
from zayt.conf.settings import Settings
from zayt.web.application import Zayt

PATH = str((Path(__file__).parent / "templates").absolute())

settings = Settings(
    default_settings
    | {
        "modules": [
            "zayt.ext.jinja",
            f"{__package__}.application",
        ],
        "jinja": {"loader": {"searchpath": [PATH]}},
    }
)


async def test_render():
    app = Zayt(settings)
    # pylint: disable=protected-access
    await app._lifespan_startup()

    client = AsyncClient(transport=ASGITransport(app=app))
    response = await client.get("http://localhost:8000/render")

    assert response.status_code == 200
    assert response.text == "Jinja"
    assert response.headers["Content-Length"] == str(len("Jinja"))
    assert "text/html" in response.headers["Content-Type"]


async def test_stream():
    app = Zayt(settings)
    # pylint: disable=protected-access
    await app._lifespan_startup()

    client = AsyncClient(transport=ASGITransport(app=app))
    response = await client.get("http://localhost:8000/stream")

    assert response.status_code == 200
    assert response.text == "Jinja"
    assert "Content-Length" not in response.headers


async def test_content_type():
    app = Zayt(settings)
    # pylint: disable=protected-access
    await app._lifespan_startup()

    client = AsyncClient(transport=ASGITransport(app=app))
    response = await client.get("http://localhost:8000/content_type")

    assert response.status_code == 200
    assert response.text == "Jinja"
    assert response.headers["Content-Length"] == str(len("Jinja"))
    assert "text/plain" in response.headers["Content-Type"]



# ruff: noqa: F401

from zayt.web.exception_handler.decorator import exception_handler
from zayt.web.lifecycle.decorator import background, startup
from zayt.web.routing.decorator import delete, get, patch, post, put, websocket



import inspect

import pytest

from zayt.util.package_scan import scan_packages


def test_scan_package():
    # pylint: disable=import-outside-toplevel
    from . import package_to_scan

    result = list(scan_packages(package_to_scan))

    # pylint: disable=import-outside-toplevel
    from .package_to_scan.module_to_scan import ClassItem, function_item

    assert result == [ClassItem, function_item]


def test_scan_package_str():
    result = list(scan_packages("tests.util.package_to_scan"))

    # pylint: disable=import-outside-toplevel
    from .package_to_scan.module_to_scan import ClassItem, function_item

    assert result == [ClassItem, function_item]


def test_scan_package_with_predicate():
    # pylint: disable=import-outside-toplevel
    from . import package_to_scan

    def predicate(arg):
        return inspect.isclass(arg)

    result = list(scan_packages(package_to_scan, predicate=predicate))

    # pylint: disable=import-outside-toplevel
    from .package_to_scan.module_to_scan import ClassItem

    assert result == [ClassItem]


def test_scan_package_str_with_predicate():
    def predicate(arg):
        return inspect.isfunction(arg)

    result = list(scan_packages("tests.util.package_to_scan", predicate=predicate))

    # pylint: disable=import-outside-toplevel
    from .package_to_scan.module_to_scan import function_item

    assert result == [function_item]


def test_non_function_predicate_should_fail():
    with pytest.raises(TypeError, match="invalid predicate"):
        list(scan_packages("", predicate="predicate"))



import asyncio
import multiprocessing as mp
from pathlib import Path
from http import HTTPStatus

from httpx import ASGITransport, AsyncClient

from .util import with_app


@with_app
async def _run(app):
    async with AsyncClient(
        transport=ASGITransport(app=app), base_url="http://server"
    ) as client:
        response = await client.get("/")
        assert response.status_code == HTTPStatus.OK
        assert response.json() == {"exception": "MyException"}


def run():
    return asyncio.run(_run())


def test_example(monkeypatch):
    path = Path(__file__).parents[2] / "examples" / "exception_handler"
    monkeypatch.syspath_prepend(path)
    monkeypatch.chdir(path)

    with mp.Pool(processes=1) as pool:
        pool.apply(run)



import inspect
from collections.abc import Awaitable, Callable
from typing import Any, ParamSpec

from asgikit.sync import run_sync

P = ParamSpec("P")


async def call_maybe_async(
    target: Awaitable | Callable[P, Any], /, *args: P.args, **kwargs: P.kwargs
) -> Any:
    if inspect.isawaitable(target):
        return await target

    if not callable(target):
        raise TypeError(f"{repr(target)} is not callable")

    call = target if inspect.isroutine(target) else getattr(target, "__call__")

    if inspect.iscoroutinefunction(call):
        return await call(*args, **kwargs)

    return await run_sync(call, *args, **kwargs)



import pytest

from zayt.di.container import Container
from zayt.di.error import ServiceNotFoundError


class Service:
    pass


def test_has_unknwon_service(ioc: Container):
    result = ioc.has(Service)
    assert not result


async def test_get_unknwon_service(ioc: Container):
    with pytest.raises(ServiceNotFoundError):
        await ioc.get(Service)


async def test_service_with_name_not_found(ioc: Container):
    ioc.register(Service, name="1")
    with pytest.raises(ServiceNotFoundError):
        await ioc.get(Service, name="2")



from typing import Annotated, Generic, TypeVar

import pytest

from zayt.di.container import Container
from zayt.di.error import TypeVarInGenericServiceError
from zayt.di.inject import Inject

T = TypeVar("T")


class GenericService(Generic[T]):
    pass


class Service(GenericService[int]):
    pass


def service_factory() -> GenericService[int]:
    return Service()


class ServiceDepends:
    service: Annotated[GenericService[int], Inject]


def service_factory_depends(dependency: GenericService[int]) -> ServiceDepends:
    instance = ServiceDepends()
    setattr(instance, "service", dependency)
    return instance


async def test_get_generic_service_with_class(ioc: Container):
    ioc.register(Service, provides=GenericService[int])

    instance = await ioc.get(GenericService[int])
    assert isinstance(instance, Service)


async def test_get_generic_service_with_factory(ioc: Container):
    ioc.register(service_factory, provides=GenericService[int])

    instance = await ioc.get(GenericService[int])
    assert isinstance(instance, Service)


async def test_get_generic_service_with_class_with_dependency(ioc: Container):
    ioc.register(Service, provides=GenericService[int])
    ioc.register(ServiceDepends)

    instance = await ioc.get(ServiceDepends)
    assert isinstance(instance, ServiceDepends)
    assert isinstance(instance.service, Service)


async def test_get_generic_service_with_factory_with_dependency(ioc: Container):
    ioc.register(service_factory, provides=GenericService[int])
    ioc.register(service_factory_depends)

    instance = await ioc.get(ServiceDepends)
    assert isinstance(instance, ServiceDepends)
    assert isinstance(instance.service, Service)


async def test_type_var_in_generic_service_should_fail(ioc: Container):
    with pytest.raises(TypeVarInGenericServiceError):
        ioc.register(Service, provides=GenericService[T])



import inspect
from collections import defaultdict
from collections.abc import AsyncGenerator, Awaitable, Generator, Iterator
from types import FunctionType
from typing import Any, TypeVar

import structlog

from asgikit.sync import run_sync

from zayt.util.maybe_async import call_maybe_async
from zayt.di.error import (
    DependencyLoopError,
    IncompatibleDependencyScopeError,
    ScopedServiceWithoutContextError,
    ServiceNotFoundError,
)
from zayt.di.interceptor import Interceptor
from zayt.di.service.model import InjectableType, ServiceDependency, ServiceSpec
from zayt.di.service.parse import parse_service_spec
from zayt.di.service.registry import ServiceRegistry

logger = structlog.get_logger(__name__)

T = TypeVar("T")


class Container:
    def __init__(self):
        self.registry = ServiceRegistry()
        self.cache: dict[int | None, dict[tuple[type, str | None], Any]] = defaultdict(
            dict
        )
        self.finalizers: dict[int | None, list[Awaitable]] = defaultdict(list)
        self.interceptors: list[type[Interceptor]] = []

    def register(
        self,
        injectable: InjectableType,
        *,
        provides: type = None,
        name: str = None,
        scoped=False,
    ):
        spec = parse_service_spec(injectable, provides, name, scoped)
        service_iface = spec.service
        service_impl = spec.impl
        name = spec.name

        self.registry[service_iface, name] = spec

        if name and service_impl:
            logger.debug(
                "service registered with implementation and name",
                service=f"{injectable.__module__}.{injectable.__qualname__}",
                implentation=f"{service_impl.__module__}.{service_impl.__qualname__}",
                name=name,
            )
        elif name:
            logger.debug(
                "service registered with name",
                service=f"{injectable.__module__}.{injectable.__qualname__}",
                name=name,
            )
        elif service_impl:
            logger.debug(
                "service registered with implementation",
                service=f"{injectable.__module__}.{injectable.__qualname__}",
                implentation=f"{service_impl.__module__}.{service_impl.__qualname__}",
            )
        else:
            logger.debug(
                "service registered",
                service=f"{injectable.__module__}.{injectable.__qualname__}",
            )

    def define(
        self,
        service_type: type,
        instance: Any,
        *,
        name: str = None,
        context: Any = None,
    ):
        cache_key = id(context) if context else None
        self.cache[cache_key][service_type, name] = instance

        if name:
            logger.debug(
                "service defined",
                service=f"{service_type.__module__}.{service_type.__qualname__}",
                name=name,
            )
        else:
            logger.debug(
                "service defined",
                service=f"{service_type.__module__}.{service_type.__qualname__}",
            )

    def interceptor(self, interceptor: type[Interceptor]):
        self.register(
            interceptor,
            provides=Interceptor,
            name=f"{interceptor.__module__}.{interceptor.__qualname__}",
        )
        self.interceptors.append(interceptor)

        logger.debug(
            "interceptor registered",
            interceptor=f"{interceptor.__module__}.{interceptor.__qualname__}",
        )

    def has(self, service_type: type, name: str = None) -> bool:
        definition = self.registry.get(service_type, name=name)
        return definition is not None

    def iter_service(
        self, service_type: type
    ) -> Iterator[tuple[type | FunctionType, str | None]]:
        record = self.registry.services.get(service_type)
        if not record:
            raise ServiceNotFoundError(service_type)

        for name, definition in record.providers.items():
            yield definition.impl, name

    def iter_all_services(
        self,
    ) -> Iterator[tuple[type, type | FunctionType | None, str | None]]:
        for _interface, record in self.registry.services.items():
            for name, definition in record.providers.items():
                yield definition.service, definition.impl, name

    async def get(
        self,
        service_type: type[T],
        *,
        name: str = None,
        optional=False,
        context: Any = None,
    ) -> T:
        dependency = ServiceDependency(service_type, name=name, optional=optional)
        return await self._get(dependency, context=context)

    async def finalize(self, context: Any = None):
        if context is not None:
            context_id = id(context)
            for finalizer in reversed(self.finalizers[context_id]):
                await finalizer

            del self.finalizers[context_id]
            return

        for context_id in self.finalizers:
            for finalizer in reversed(self.finalizers[context_id]):
                await finalizer

    def _get_from_cache(
        self,
        service_type: type[T],
        name: str | None,
        context: Any = None,
    ) -> T | None:
        if context is not None:
            context_id = id(context)
            if instance := self.cache[context_id].get((service_type, name)):
                return instance

        return self.cache[None].get((service_type, name))

    async def _get(
        self,
        dependency: ServiceDependency,
        context: Any = None,
        stack: list[tuple[type, str | None]] = None,
    ) -> Any | None:
        service_type, service_name, optional = (
            dependency.service,
            dependency.name,
            dependency.optional,
        )

        # check if service exists in cache
        if instance := self._get_from_cache(service_type, service_name, context):
            return instance

        try:
            service_spec = self.registry.get(service_type, service_name)
            if not service_spec:
                raise ServiceNotFoundError(service_type, service_name)
        except ServiceNotFoundError:
            if optional:
                return None
            raise

        stack = stack or []

        if stack:
            parent_service, parent_name = stack[-1]
            parent_spec = self.registry.get(parent_service, parent_name)
            if service_spec.scoped and not parent_spec.scoped:
                raise IncompatibleDependencyScopeError(
                    parent_service, service_type, parent_name
                )

        if (service_type, service_name) in stack:
            raise DependencyLoopError(stack, (service_type, service_name))

        stack.append((service_type, service_name))
        instance = await self._create_service(service_spec, context, stack)
        stack.pop()

        return instance

    async def _get_dependent_services(
        self,
        service_spec: ServiceSpec,
        context: Any | None,
        stack: list,
    ) -> dict[str, Any]:
        return {
            name: await self._get(dep, context, stack)
            for name, dep in service_spec.dependencies
        }

    async def _create_service(
        self,
        service_spec: ServiceSpec,
        context: Any | None,
        stack: list[tuple[type[T], str]],
    ) -> T:
        name = service_spec.name

        if not service_spec.scoped:
            context = None
        elif context is None:
            raise ScopedServiceWithoutContextError(service_spec.service)

        context_id = id(context) if context else None

        if factory := service_spec.factory:
            dependencies = await self._get_dependent_services(
                service_spec, context, stack
            )

            instance = await call_maybe_async(factory, **dependencies)
            if inspect.isgenerator(instance):
                generator = instance
                instance = await run_sync(next, generator)
                self._setup_generator_finalizer(generator, context)
            elif inspect.isasyncgen(instance):
                generator = instance
                instance = await anext(generator)
                self._setup_asyncgen_finalizer(generator, context)

            self.cache[context_id][service_spec.service, name] = instance
        else:
            instance = service_spec.impl()
            self.cache[context_id][service_spec.service, name] = instance

            dependencies = await self._get_dependent_services(
                service_spec, context, stack
            )

            for name, dep_service in dependencies.items():
                setattr(instance, name, dep_service)

            if initializer := service_spec.initializer:
                await call_maybe_async(initializer, instance)

            self._setup_finalizer(service_spec, instance, context_id)

        if service_spec.service is not Interceptor:
            await self._run_interceptors(instance, service_spec.service)

        return instance

    def _setup_finalizer(
        self, service_spec: ServiceSpec, instance: Any, context_id: int | None
    ):
        if finalizer := service_spec.finalizer:
            self.finalizers[context_id].append(call_maybe_async(finalizer, instance))

    def _setup_generator_finalizer(self, gen: Generator, context: Any | None):
        context_id = id(context) if context else None
        self.finalizers[context_id].append(run_sync(next, gen, None))

    def _setup_asyncgen_finalizer(self, gen: AsyncGenerator, context: Any | None):
        context_id = id(context) if context else None
        self.finalizers[context_id].append(anext(gen, None))

    async def _run_interceptors(self, instance: Any, service_type: type):
        for cls in self.interceptors:
            interceptor = await self.get(
                Interceptor, name=f"{cls.__module__}.{cls.__qualname__}"
            )
            await call_maybe_async(interceptor.intercept, instance, service_type)



from importlib import import_module
from pkgutil import resolve_name

__all__ = ("import_item",)


def import_item(name: str):
    """Import a module or an item within a module using its name

    :param name: The name of the item to import, in the format 'package.module[:item]'.

    :return: The imported module or item within the module
    """

    if ":" in name:
        return resolve_name(name)

    try:
        return import_module(name)
    except ImportError as err:
        match name.rsplit(".", maxsplit=1):
            case [mod, item]:
                return resolve_name(f"{mod}:{item}")
            case _:
                raise err



import pytest

from zayt.di.container import Container
from zayt.di.error import FactoryMissingReturnTypeError


class Service1:
    pass


def service1_factory() -> Service1:
    yield Service1()


class Service2:
    def __init__(self, service1: Service1):
        self.service1 = service1


def service2_factory(service1: Service1) -> Service2:
    yield Service2(service1)


class Interface:
    pass


class Implementation(Interface):
    pass


def interface_factory() -> Interface:
    yield Implementation()


def test_has_service(ioc: Container):
    ioc.register(service1_factory)
    assert ioc.has(Service1)


async def test_service_with_provided_interface(ioc: Container):
    ioc.register(interface_factory)

    instance = await ioc.get(Interface)
    assert isinstance(instance, Implementation)
    await ioc.finalize()


async def test_inject_singleton(ioc: Container):
    ioc.register(service1_factory)
    ioc.register(service2_factory)

    instance = await ioc.get(Service2)
    assert isinstance(instance, Service2)
    assert isinstance(instance.service1, Service1)

    other_instance = await ioc.get(Service2)
    assert other_instance is instance
    assert other_instance.service1 is instance.service1

    await ioc.finalize()


def test_factory_function_without_return_type_should_fail(ioc: Container):
    async def service_factory():
        pass

    with pytest.raises(FactoryMissingReturnTypeError):
        ioc.register(service_factory)


def test_provides_option_should_log_warning(ioc: Container, log_output):
    ioc.register(interface_factory, provides=Interface)

    assert (
        log_output.entries[0]["event"]
        == "option 'provides' on a factory function has no effect"
    )
    assert (
        log_output.entries[0]["service"]
        == "tests.di.test_service_generator.interface_factory"
    )



from zayt.conf.defaults import default_settings
from zayt.conf.settings import Settings
from zayt.web.application import Zayt
from zayt.web.lifecycle.decorator import startup


@startup
def on_startup():
    print("startup", end="")


async def test_application(capfd):
    settings = Settings(
        default_settings
        | {
            "modules": [f"{test_application.__module__}"],
        }
    )

    app = Zayt(settings)

    # pylint: disable=protected-access
    await app._lifespan_startup()
    assert capfd.readouterr().out == "startup"



from typing import Annotated

from zayt.di.container import Container
from zayt.di.inject import Inject


class DependentService:
    pass


class ServiceWithOptionalDep:
    dependent: Annotated[DependentService, Inject] = None


async def test_optional_dependency(ioc: Container):
    ioc.register(ServiceWithOptionalDep)
    instance = await ioc.get(ServiceWithOptionalDep)
    assert instance.dependent is None


async def test_optional_dependency_with_provided_dependency(ioc: Container):
    ioc.register(ServiceWithOptionalDep)
    ioc.register(DependentService)
    instance = await ioc.get(ServiceWithOptionalDep)
    assert isinstance(instance.dependent, DependentService)



from http import HTTPStatus

from httpx import ASGITransport, AsyncClient

from zayt.conf.defaults import default_settings
from zayt.conf.settings import Settings
from zayt.web.application import Zayt


async def test_application():
    settings = Settings(default_settings | {"modules": [f"{__package__}.application"]})
    app = Zayt(settings)

    client = AsyncClient(transport=ASGITransport(app=app))
    response = await client.get("http://localhost:8000/")
    assert response.text == "Ok"


async def test_not_found():
    settings = Settings(default_settings | {"modules": [f"{__package__}.application"]})
    app = Zayt(settings)

    client = AsyncClient(transport=ASGITransport(app=app))
    response = await client.get("http://localhost:8000/not-found")
    assert response.status_code == HTTPStatus.NOT_FOUND



import logging
import logging.config
import sys

import structlog

from zayt.conf.settings import Settings


def setup(settings: Settings):
    root_level = settings.logging.get("root", "WARN").upper()
    log_format = settings.logging.get(
        "format", "console" if sys.stderr.isatty() else "logfmt"
    )

    match log_format:
        case "console":
            renderer = structlog.dev.ConsoleRenderer(colors=True)
        case "json":
            renderer = structlog.processors.JSONRenderer()
        case "logfmt":
            renderer = structlog.processors.LogfmtRenderer(
                key_order=["timestamp", "level", "event"]
            )
        case "keyvalue":
            renderer = structlog.processors.KeyValueRenderer(
                key_order=["timestamp", "level", "event"]
            )
        case _:
            raise ValueError("Invalid log format")

    timestamper = structlog.processors.TimeStamper(
        fmt="%Y-%m-%d %H:%M:%S" if log_format == "console" else "iso"
    )

    shared_processors = [
        structlog.contextvars.merge_contextvars,
        timestamper,
        structlog.stdlib.add_logger_name,
        structlog.stdlib.add_log_level,
        structlog.stdlib.PositionalArgumentsFormatter(),
        structlog.processors.StackInfoRenderer(),
        structlog.processors.format_exc_info,
    ]

    extra_loggers = {
        "sqlalchemy.engine.Engine": {
            "level": root_level,
            "handlers": ["console"],
            "propagate": False,
        },
        "uvicorn": {
            "handlers": [],
            "propagate": True,
        },
        "_granian": {
            "handlers": [],
            "propagate": True,
        },
        "granian.access": {
            "handlers": [],
            "propagate": True,
        },
    }

    loggers = {
        module: {"level": level.upper(), "handlers": ["console"], "propagate": False}
        for module, level in settings.logging.get("level", {}).items()
    }

    logging.config.dictConfig(
        {
            "version": 1,
            "disable_existing_loggers": False,
            "formatters": {
                "default": {
                    "()": structlog.stdlib.ProcessorFormatter,
                    "processors": [
                        structlog.stdlib.ProcessorFormatter.remove_processors_meta,
                        renderer,
                    ],
                    "foreign_pre_chain": shared_processors,
                }
            },
            "handlers": {
                "console": {
                    "class": "logging.StreamHandler",
                    "formatter": "default",
                }
            },
            "root": {
                "handlers": ["console"],
                "level": root_level,
            },
            "loggers": extra_loggers | loggers,
        }
    )

    structlog.configure(
        processors=[
            *shared_processors,
            structlog.stdlib.ProcessorFormatter.wrap_for_formatter,
        ],
        logger_factory=structlog.stdlib.LoggerFactory(),
        wrapper_class=structlog.stdlib.BoundLogger,
        cache_logger_on_first_use=True,
    )


# class DevFormatter(logging.Formatter):
#     def __init__(self, *args, **kwargs):
#         super().__init__(datefmt="%Y-%m-%d %H:%M:%S")
#
#     def format(self, record: logging.LogRecord) -> str:
#         message = record.getMessage()
#
#         if ctx := getattr(record, "ctx", None):
#             values = {}
#             for key, value in ctx.items():
#                 if isinstance(value, (int, float, bool, str)):
#                     values[key] = value
#                 elif isinstance(value, (date, time, datetime)):
#                     values[key] = value.isoformat()
#                 elif isinstance(value, Decimal):
#                     values[key] = str(value)
#                 else:
#                     values[key] = json.dumps(str(value))
#
#             context = " ".join(
#                 f"{name}={value}"
#                 for name, value in values.items()
#             )
#         else:
#             context = None
#
#         record.asctime = self.formatTime(record, self.datefmt)
#
#         if context:
#             s = f"{record.asctime} {record.levelname:8} {message:45} [{record.name}] {context}"
#         else:
#             s = f"{record.asctime} {record.levelname:8} {message}"
#
#         if record.exc_info:
#             # Cache the traceback text to avoid converting it multiple times
#             # (it's constant anyway)
#             if not record.exc_text:
#                 record.exc_text = self.formatException(record.exc_info)
#         if record.exc_text:
#             if s[-1:] != "\n":
#                 s = s + "\n"
#             s = s + record.exc_text
#         if record.stack_info:
#             if s[-1:] != "\n":
#                 s = s + "\n"
#             s = s + self.formatStack(record.stack_info)
#
#         return s
#
#
# class JsonFormatter(logging.Formatter):
#     def format(self, record: logging.LogRecord) -> str:
#         values = {
#             "time": datetime.fromtimestamp(record.created).isoformat(),
#             "level": record.levelname,
#             "event": record.getMessage(),
#             "source": record.name,
#         }
#
#         if ctx := getattr(record, "ctx", None):
#             for key, value in ctx.items():
#                 if isinstance(value, (int, float, bool, str)):
#                     values[key] = value
#                 elif isinstance(value, (date, time, datetime)):
#                     values[key] = value.isoformat()
#                 else:
#                     values[key] = str(value)
#
#         if record.exc_info:
#             # Cache the traceback text to avoid converting it multiple times
#             # (it's constant anyway)
#             if not record.exc_text:
#                 record.exc_text = self.formatException(record.exc_info)
#
#         if record.exc_text:
#             values["exception"] = record.exc_text
#
#         if record.stack_info:
#             values["stack"] = self.formatStack(record.stack_info)
#
#         return json.dumps(values)



from typing import Annotated

import pytest

from zayt.di.container import Container
from zayt.di.error import IncompatibleDependencyScopeError
from zayt.di.inject import Inject


class Service1:
    pass


class Service2:
    service1: Annotated[Service1, Inject]


class Service3:
    service2: Annotated[Service2, Inject]


async def test_service_chain_with_mixed_scopes(ioc: Container):
    ioc.register(Service1, scoped=True)
    ioc.register(Service2, scoped=False)
    ioc.register(Service3, scoped=True)

    context = object()

    with pytest.raises(IncompatibleDependencyScopeError):
        await ioc.get(Service3, context=context)



import structlog
from sqlalchemy.engine.url import URL, make_url
from sqlalchemy.ext.asyncio import (
    AsyncEngine,
    AsyncSession,
    async_sessionmaker,
    create_async_engine,
)

from zayt.util.import_item import import_item
from zayt.conf.settings import Settings
from zayt.di.container import Container
from zayt.di.decorator import service

logger = structlog.get_logger(__name__)


def _get_url(settings: Settings) -> URL:
    if url := settings.get("url"):
        url = make_url(url)
        if username := settings.get("username"):
            url = url.set(username=username)
        if password := settings.get("password"):
            url = url.set(password=password)
        if query := settings.get("query"):
            url = url.set(query=query)
    else:
        kwargs = {}

        if drivername := settings.get("drivername"):
            kwargs["drivername"] = drivername
        if host := settings.get("host"):
            kwargs["host"] = host
        if port := settings.get("port"):
            kwargs["port"] = int(port)
        if database := settings.get("database"):
            kwargs["database"] = database
        if username := settings.get("username"):
            kwargs["username"] = username
        if password := settings.get("password"):
            kwargs["password"] = password
        if query := settings.get("query"):
            kwargs["query"] = query

        url = URL.create(**kwargs)

    return url


def make_engine_service(name: str):
    async def engine_service(settings: Settings) -> AsyncEngine:
        sa_settings = settings.sqlalchemy.connections[name]

        url = _get_url(sa_settings)
        options = dict(sa_settings.get("options", {}))

        engine = create_async_engine(url, **options)
        yield engine
        await engine.dispose()

    return engine_service


@service
async def engine_dict_service(
    settings: Settings, di: Container
) -> dict[str, AsyncEngine]:
    return {
        db: await di.get(AsyncEngine, name=db if db else None)
        for db in settings.sqlalchemy.connections
    }


@service
async def sessionmaker_service(
    settings: Settings, engines_map: dict[str, AsyncEngine]
) -> async_sessionmaker:
    session_settings = settings.sqlalchemy.get("session", {})

    args = []
    kwargs = dict(session_settings.get("options", {}))

    if binds := session_settings.get("binds"):
        binds_config = {}
        for mapper, engine_name in binds.items():
            if isinstance(mapper, str):
                mapper = import_item(mapper)
            if engine := engines_map.get(engine_name):
                binds_config[mapper] = engine
            else:
                raise ValueError(f"No engine with name '{engine_name}'")

        kwargs["binds"] = binds_config
    else:
        engine = engines_map.get("")
        if not engine:
            name, engine = next(iter(engines_map.items()))
            logger.warning(
                "No default engine found, using next available",
                engine=name,
            )
        args.append(engine)

    if class_ := kwargs.pop("class", None):
        kwargs["class_"] = class_

    return async_sessionmaker(*args, **kwargs)


@service(scoped=True)
async def async_session(sessionmaker: async_sessionmaker) -> AsyncSession:
    async with sessionmaker() as session:
        yield session



import pytest

from zayt.util.import_item import import_item


@pytest.mark.parametrize(
    "separator",
    [".", ":"],
    ids=["dot", "colon"],
)
def test_import_item(separator):
    item = import_item(f"tests.util.package_to_scan.module_to_scan{separator}ClassItem")

    # pylint: disable=import-outside-toplevel
    from .package_to_scan.module_to_scan import ClassItem

    assert item is ClassItem


@pytest.mark.parametrize(
    "name",
    [
        "invalid",
        "invalid.item",
        "invalid:item",
        "invalid.module.item",
        "invalid.module:item",
    ],
)
def test_import_item_invalid_name_should_fail(name):
    with pytest.raises(ImportError):
        import_item(name)


@pytest.mark.parametrize(
    "separator",
    [".", ":"],
    ids=["dot", "colon"],
)
def test_import_missing_item_from_module_should_fail(separator):
    with pytest.raises(AttributeError):
        import_item(
            f"tests.util.package_to_scan.module_to_scan{separator}does_not_exist"
        )



import pytest

from zayt.di.error import InvalidServiceTypeError
from zayt.di.service.parse import _get_service_signature, parse_service_spec


def test_parse_invalid_service_should_fail():
    with pytest.raises(InvalidServiceTypeError):
        parse_service_spec(object())


def test_get_dependencies_invalid_service_should_fail():
    with pytest.raises(InvalidServiceTypeError):
        list(_get_service_signature(object()))



import inspect

ATTRIBUTE_STARTUP = "__zayt_startup__"
ATTRIBUTE_BACKGROUND = "__zayt_background__"


def startup(target):
    assert inspect.isfunction(target)
    setattr(target, ATTRIBUTE_STARTUP, True)
    return target


def background(target):
    assert inspect.iscoroutinefunction(target)
    setattr(target, ATTRIBUTE_BACKGROUND, True)
    return target



default_settings = {
    "modules": [],
    "asgi_middleware": [],
    "logging": {
        "setup": "zayt.logging:setup",
    },
    "jinja": {},
    "memcached": {},
    "redis": {},
    "sqlalchemy": {
        "connections": {},
        "session": {},
    },
    "staticfiles": {
        "path": "/static",
        "root": "resources/static",
        "mappings": {},
    },
    "uploadedfiles": {
        "path": "/uploads",
        "root": "resources/uploads",
    },
}



import os
from http import HTTPStatus
from pathlib import Path
from typing import Annotated, Any

from asgikit import Request
from jinja2 import Environment, FileSystemLoader

from zayt.conf import Settings
from zayt.di.decorator import service
from zayt.di.inject import Inject


@service
def jinja_environment(settings: Settings) -> Environment:
    jinja_settings = settings.jinja

    kwargs: dict[str, Any] = dict(jinja_settings)

    if loader_settings := jinja_settings.get("loader"):
        if isinstance(loader_settings, Settings):
            loader_settings = dict(loader_settings)

            if "searchpath" not in loader_settings:
                loader_settings["searchpath"] = [
                    Path(os.getcwd()) / "resources" / "templates"
                ]

            if not isinstance(loader_settings["searchpath"], list):
                loader_settings["searchpath"] = [loader_settings["searchpath"]]

            kwargs["loader"] = FileSystemLoader(**loader_settings)
    else:
        kwargs["loader"] = FileSystemLoader(
            searchpath=Path(os.getcwd()) / "resources" / "templates"
        )

    kwargs.pop("enable_async", None)

    return Environment(enable_async=True, **kwargs)


@service
class JinjaTemplate:
    environment: Annotated[Environment, Inject]

    async def render(
        self,
        template_name: str,
        context: dict,
    ):
        template = self.environment.get_template(template_name)
        return await template.render_async(context)

    # pylint: disable=too-many-arguments
    async def respond(
        self,
        request: Request,
        template_name: str,
        context: dict,
        *,
        status=HTTPStatus.OK,
        media_type="text/html",
        stream: bool = False,
    ):
        template = self.environment.get_template(template_name)

        if stream:
            render_stream = template.generate_async(context)
            await request.respond_stream(
                (i.encode() async for i in render_stream),
                status=status,
                media_type=media_type,
            )
        else:
            rendered = await template.render_async(context)
            await request.respond_text(
                rendered,
                status=status,
                media_type=media_type,
            )



from zayt.di.container import Container


class Service:
    pass


async def test_define_service_not_registered(ioc: Container):
    instance = Service()
    ioc.define(Service, instance)

    result = await ioc.get(Service)

    assert result is instance


async def test_define_service_already_registered(ioc: Container):
    ioc.register(Service)
    instance = Service()
    ioc.define(Service, instance)

    result = await ioc.get(Service)

    assert result is instance



import inspect

from zayt.conf.settings import get_settings
from zayt.web.application import Zayt


def with_app(target):
    assert inspect.iscoroutinefunction(target)

    async def inner(*args, **kwargs):
        app = Zayt(get_settings())
        await app._lifespan_startup()

        try:
            await target(app, *args, **kwargs)
        finally:
            await app._lifespan_shutdown()

    return inner



from zayt.web.exception_handler.decorator import exception_handler
from zayt.web.routing.decorator import get


class MyException(Exception):
    pass


class MyBaseException(Exception):
    pass


class DerivedException(MyBaseException):
    pass


@exception_handler(MyException)
async def my_exception_handler(request, exc):
    await request.respond_json({"exception": exc.__class__.__name__})


@exception_handler(BaseException)
async def base_exception_handler(request, exc):
    await request.respond_text(f"handler=base; exception={exc.__class__.__name__}")


@get
async def index(request):
    raise MyException()


@get("base")
async def base(request):
    raise MyBaseException()


@get("derived")
async def derived(request):
    raise DerivedException()



from asgikit import Request
from sqlalchemy import text
from sqlalchemy.ext.asyncio import async_sessionmaker

from zayt.web import get


@get
async def index(request: Request, sessionmaker: async_sessionmaker):
    async with sessionmaker() as session:
        result = await session.execute(text("select sqlite_version()"))
        version = result.first()[0]

    await request.respond_json(version)



import inspect
import typing
from collections.abc import Callable, Iterator
from typing import Annotated, Any, TypeVar

import structlog

from zayt.di.error import (
    FactoryMissingReturnTypeError,
    InvalidDependencyAnnotationError,
    InvalidServiceTypeError,
    ServiceWithUntypedDependencyError,
    TypeVarInGenericServiceError,
)
from zayt.di.inject import Inject
from zayt.di.service.model import InjectableType, ServiceDependency, ServiceSpec

logger = structlog.get_logger(__name__)

DI_INITIALIZER = "initialize"
DI_FINALIZER = "finalize"


def parse_service_spec(
    injectable: InjectableType,
    provides: type = None,
    name: str = None,
    scoped=False,
) -> ServiceSpec:
    if not (inspect.isclass(injectable) or inspect.isfunction(injectable)):
        raise InvalidServiceTypeError(injectable)

    if inspect.isclass(injectable):
        provided_service, initializer, finalizer = _parse_definition_class(
            injectable, provides
        )

        implementation = injectable
        factory = None
    elif inspect.isfunction(injectable):
        if provides:
            logger.warning(
                "option 'provides' on a factory function has no effect",
                service=f"{injectable.__module__}.{injectable.__name__}",
            )

        provided_service = _parse_definition_factory(injectable)
        initializer = None
        finalizer = None

        implementation = None
        factory = injectable
    else:
        raise RuntimeError()

    dependencies = list(get_dependencies(injectable))

    return ServiceSpec(
        service=provided_service,
        impl=implementation,
        factory=factory,
        name=name,
        scoped=scoped,
        dependencies=dependencies,
        initializer=initializer,
        finalizer=finalizer,
    )


def get_dependencies(
    service: InjectableType,
    skip=0,
) -> Iterator[tuple[str, ServiceDependency]]:
    for name, hint, meta, is_optional in _get_service_signature(service, skip=skip):
        if isinstance(meta, Inject):
            service_name = meta.name
        else:
            service_name = None

        dependency = ServiceDependency(hint, name=service_name, optional=is_optional)
        yield name, dependency


def _get_service_signature(
    service: InjectableType,
    *,
    skip=0,
) -> Iterator[tuple[str, type, Any, bool]]:
    if inspect.isclass(service):
        for name, hint in typing.get_type_hints(service, include_extras=True).items():
            try:
                params = _get_injectable_params(hint)
            except InvalidDependencyAnnotationError:
                continue

            if not params:
                continue

            arg_type, arg_meta = params
            is_optional = hasattr(service, name)
            yield name, arg_type, arg_meta, is_optional

    elif inspect.isfunction(service):
        skipped = 0
        for name, param in inspect.signature(service).parameters.items():
            if skipped < skip:
                skipped += 1
                continue

            hint = param.annotation
            if hint is inspect.Signature.empty:
                raise ServiceWithUntypedDependencyError(service, name)

            is_optional = param.default is not inspect.Parameter.empty
            if params := _get_injectable_params(hint):
                arg_type, arg_meta = params
                yield name, arg_type, arg_meta, is_optional
            else:
                yield name, hint, None, is_optional

    else:
        raise InvalidServiceTypeError(service)


def _get_injectable_params(hint) -> tuple[type, Any] | None:
    if typing.get_origin(hint) is not Annotated:
        return None

    arg_type, arg_meta = typing.get_args(hint)[0:2]
    if not isinstance(arg_meta, Inject) and arg_meta is not Inject:
        raise InvalidDependencyAnnotationError(arg_type, arg_meta)

    return arg_type, arg_meta


def _parse_definition_class(
    service: type, provides: type | None
) -> tuple[type, Callable, Callable]:
    if provides and typing.get_origin(provides):
        if any(isinstance(a, TypeVar) for a in typing.get_args(provides)):
            raise TypeVarInGenericServiceError(provides)

    provided_service = provides or service

    initializer = getattr(service, DI_INITIALIZER, None)
    finalizer = getattr(service, DI_FINALIZER, None)

    return provided_service, initializer, finalizer


def _parse_definition_factory(service: Callable) -> type:
    service_type = typing.get_type_hints(service).get("return")
    if service_type is None:
        raise FactoryMissingReturnTypeError(service)

    provided_service = service_type

    return provided_service



from typing import Annotated

from asgikit import Request
from sqlalchemy import text
from sqlalchemy.ext.asyncio import AsyncEngine

from zayt.di import Inject
from zayt.web import get


@get
async def index(request: Request, engine: Annotated[AsyncEngine, Inject(name="other")]):
    async with engine.begin() as conn:
        result = await conn.execute(text("select sqlite_version()"))
        version = result.first()[0]

    await request.respond_text(version)



import os
from pathlib import Path

from zayt.util import dotenv


def test_dotenv(monkeypatch):
    monkeypatch.chdir(Path(__file__).parent / "dotenv")
    dotenv.init()

    assert "TEST_VARIABLE" in os.environ
    assert os.environ["TEST_VARIABLE"] == "value"


def test_dotenv_custom_location(monkeypatch):
    monkeypatch.setenv("ZAYT_DOTENV", str(Path(__file__).parent / "dotenv" / ".env"))
    dotenv.init()

    assert "TEST_VARIABLE" in os.environ
    assert os.environ["TEST_VARIABLE"] == "value"



from http import HTTPStatus

import pytest
from httpx import ASGITransport, AsyncClient

from zayt.conf.defaults import default_settings
from zayt.conf.settings import Settings
from zayt.web.application import Zayt


@pytest.mark.parametrize(
    "application,database",
    [
        ("application", ""),
        ("application_named", "other"),
    ],
    ids=["default", "named"],
)
async def test_application(application: str, database: str):
    settings = Settings(
        default_settings
        | {
            "modules": [
                "zayt.ext.sqlalchemy",
                f"{__package__}.{application}",
            ],
            "sqlalchemy": {
                "connections": {database: {"url": "sqlite+aiosqlite:///:memory:"}}
            },
        }
    )

    app = Zayt(settings)

    # pylint: disable=protected-access
    await app._lifespan_startup()

    client = AsyncClient(transport=ASGITransport(app=app))
    response = await client.get("http://localhost:8000/")

    assert response.status_code == HTTPStatus.OK, response.text



from typing import Annotated

import pytest

from zayt.di.container import Container
from zayt.di.error import ServiceAlreadyRegisteredError, ServiceNotFoundError
from zayt.di.inject import Inject


class NamedService:
    pass


class ServiceWithNamedDep:
    dependent: Annotated[NamedService, Inject("1")]


class Interface:
    pass


class Impl1(Interface):
    pass


class Impl2(Interface):
    pass


async def test_dependency_with_name(ioc: Container):
    ioc.register(NamedService, name="1")
    ioc.register(ServiceWithNamedDep)

    instance = await ioc.get(ServiceWithNamedDep)
    dependent = instance.dependent
    assert isinstance(dependent, NamedService)


async def test_unnamed_dependency_with_named_service_should_fail(ioc: Container):
    class Service:
        pass

    class DependentService:
        dependent: Annotated[Service, Inject("1")]

    ioc.register(NamedService)
    ioc.register(DependentService)

    with pytest.raises(ServiceNotFoundError):
        await ioc.get(DependentService)


async def test_named_dependency_with_unnamed_service_should_fail(ioc: Container):
    class Service:
        pass

    class DependentService:
        dependent: Annotated[Service, Inject]

    ioc.register(NamedService, name="1")
    ioc.register(DependentService)

    with pytest.raises(ServiceNotFoundError):
        await ioc.get(DependentService)


async def test_register_two_services_with_the_same_name_should_fail(ioc: Container):
    ioc.register(NamedService)
    with pytest.raises(ServiceAlreadyRegisteredError):
        ioc.register(NamedService)


async def test_dependencies_with_same_interface(ioc: Container):
    ioc.register(Impl1, provides=Interface, name="impl1")
    ioc.register(Impl2, provides=Interface, name="impl2")

    service1 = await ioc.get(Interface, name="impl1")
    service2 = await ioc.get(Interface, name="impl2")

    assert service1 is not service2



from typing import Annotated

import pytest

from zayt.di.container import Container
from zayt.di.inject import Inject


class ServiceInitialize:
    def initialize(self):
        pass


class ServiceAsyncInitialize:
    async def initialize(self):
        pass


class Service:
    def finalization_method(self):
        pass


class ServiceFinalize:
    def finalize(self):
        pass


class ServiceAsyncFinalize:
    async def finalize(self):
        pass


def factory_finalize() -> Service:
    instance = Service()
    yield instance
    instance.finalization_method()


async def factory_async_finalize() -> Service:
    instance = Service()
    yield instance
    instance.finalization_method()


class FinalizerOrder1:
    def initialize(self):
        pass

    def finalize(self):
        pass


class FinalizerOrder2:
    dep: Annotated[FinalizerOrder1, Inject]

    def initialize(self):
        pass

    def finalize(self):
        pass


class ContextScopeFinalize:
    async def finalize(self):
        pass


@pytest.mark.parametrize("service_class", [ServiceInitialize, ServiceAsyncInitialize])
async def test_call_initialize(ioc: Container, mocker, service_class):
    initializer = mocker.patch.object(service_class, "initialize")
    ioc.register(service_class)

    await ioc.get(service_class)
    initializer.assert_called_once()


@pytest.mark.parametrize("service_class", [ServiceFinalize, ServiceAsyncFinalize])
async def test_call_finalize(ioc: Container, mocker, service_class):
    finalizer = mocker.patch.object(service_class, "finalize")
    ioc.register(service_class)

    await ioc.get(service_class)
    await ioc.finalize()

    finalizer.assert_called_once()


@pytest.mark.parametrize("factory", [factory_finalize, factory_async_finalize])
async def test_call_factory_finalize(ioc: Container, mocker, factory):
    finalizer = mocker.patch.object(Service, "finalization_method")
    ioc.register(factory)

    await ioc.get(Service)
    await ioc.finalize()

    finalizer.assert_called_once()


async def test_finalizer_order(ioc: Container, mocker):
    output = []

    def new_method(self):
        output.append(self.__class__.__name__)

    mocker.patch.object(FinalizerOrder1, "initialize", new_method)
    mocker.patch.object(FinalizerOrder2, "initialize", new_method)
    mocker.patch.object(FinalizerOrder1, "finalize", new_method)
    mocker.patch.object(FinalizerOrder2, "finalize", new_method)

    ioc.register(FinalizerOrder1)
    ioc.register(FinalizerOrder2)

    await ioc.get(FinalizerOrder2)
    await ioc.finalize()

    expected = [
        "FinalizerOrder1",
        "FinalizerOrder2",
        "FinalizerOrder2",
        "FinalizerOrder1",
    ]
    assert output == expected


async def test_context_finalizer_is_called(ioc: Container, mocker):
    service_finalize_func = mocker.patch.object(ServiceFinalize, "finalize")
    context_scope_finalize_func = mocker.patch.object(ContextScopeFinalize, "finalize")

    ioc.register(ServiceFinalize)
    ioc.register(ContextScopeFinalize, scoped=True)

    context = object()

    await ioc.get(ServiceFinalize, context=context)
    await ioc.get(ContextScopeFinalize, context=context)

    await ioc.finalize(context)

    service_finalize_func.assert_not_called()
    context_scope_finalize_func.assert_called_once()

    await ioc.finalize()
    service_finalize_func.assert_called_once()


async def test_all_finalizers_called(ioc: Container, mocker):
    service_finalize_func = mocker.patch.object(ServiceFinalize, "finalize")
    context_scope_finalize_func = mocker.patch.object(ContextScopeFinalize, "finalize")

    ioc.register(ServiceFinalize)
    ioc.register(ContextScopeFinalize)

    context = object()

    await ioc.get(ServiceFinalize)
    await ioc.get(ContextScopeFinalize, context=context)

    await ioc.finalize()
    service_finalize_func.assert_called_once()
    context_scope_finalize_func.assert_called_once()



import pytest

from zayt.util.maybe_async import call_maybe_async


async def test_awaitable_object():
    async def func():
        return 1

    awaitable = func()
    result = await call_maybe_async(awaitable)
    assert result == 1


async def test_not_callable_should_fail():
    not_callable = None

    with pytest.raises(TypeError):
        await call_maybe_async(not_callable)


async def test_call_async_function():
    async def func():
        return 1

    result = await call_maybe_async(func)
    assert result == 1


async def test_call_async_function_with_args():
    async def func(param: int):
        return param

    result = await call_maybe_async(func, 1)
    assert result == 1


async def test_call_async_function_with_kwargs():
    async def func(param: int):
        return param

    result = await call_maybe_async(func, param=1)
    assert result == 1


async def test_call_async_method():
    class Klass:
        async def func(self):
            return 1

    obj = Klass()

    result = await call_maybe_async(obj.func)
    assert result == 1


async def test_call_async_method_with_args():
    class Klass:
        async def func(self, param: int):
            return param

    obj = Klass()

    result = await call_maybe_async(obj.func, 1)
    assert result == 1


async def test_call_async_method_with_kwargs():
    class Klass:
        async def func(self, param: int):
            return param

    obj = Klass()

    result = await call_maybe_async(obj.func, param=1)
    assert result == 1


async def test_call_async_object():
    class Klass:
        async def __call__(self):
            return 1

    obj = Klass()

    result = await call_maybe_async(obj)
    assert result == 1


async def test_call_async_object_with_args():
    class Klass:
        async def __call__(self, param: int):
            return param

    obj = Klass()

    result = await call_maybe_async(obj, 1)
    assert result == 1


async def test_call_async_object_with_kwargs():
    class Klass:
        async def __call__(self, param: int):
            return param

    obj = Klass()

    result = await call_maybe_async(obj, param=1)
    assert result == 1



import os
from abc import ABC, abstractmethod
from collections.abc import Callable
from pathlib import Path

from asgikit import Request

from zayt.conf import Settings
from zayt.web.exception import HTTPNotFoundException


def static_files_middleware(app, settings: Settings):
    settings = settings.staticfiles
    path = settings.path.lstrip("/")
    root = Path(settings.root).resolve().absolute()

    # dict of file path to (content-type, content-length, last-modified)
    filelist: dict[str, os.stat_result] = {}

    for dirpath, _dirnames, filenames in os.walk(root):
        for filename in filenames:
            file = os.path.join(dirpath, filename)
            stat_result = os.stat(file)
            filelist[file] = stat_result

    mappings = {
        name.lstrip("/"): os.path.join(root, value.lstrip("/"))
        for name, value in settings.get("mappings", {}).items()
    }

    if difference := set(mappings.values()).difference(filelist):
        files = ", ".join(difference)
        raise ValueError(f"Static files mappings not found: {files}")

    return StaticFilesMiddleware(app, path, root, filelist, mappings)


def uploaded_files_middleware(app, settings: Settings):
    settings = settings.uploadedfiles
    path = settings.path.lstrip("/")
    root = Path(settings.root).resolve()
    return UploadedFilesMiddleware(app, path, root)


class BaseFilesMiddleware(ABC):
    def __init__(self, app: Callable, path: str, root: Path):
        self.app = app
        self.path = path if path.endswith("/") else path + "/"
        self.root = root

    @abstractmethod
    def get_file_to_serve(
        self, scope: dict
    ) -> tuple[str, os.stat_result | None] | None:
        pass

    async def __call__(self, scope, receive, send):
        if (scope["type"] == "http") and (result := self.get_file_to_serve(scope)):
            file_to_serve, stat_result = result
            file_to_serve = (self.root / file_to_serve).resolve()

            if not (
                file_to_serve.is_file() and file_to_serve.is_relative_to(self.root)
            ):
                raise HTTPNotFoundException()

            request = Request(scope, receive, send)
            await request.respond_file(file_to_serve, stat_result=stat_result)
        else:
            await self.app(scope, receive, send)


class UploadedFilesMiddleware(BaseFilesMiddleware):
    def get_file_to_serve(self, scope: dict) -> tuple[str, None] | None:
        request_path = scope["path"].lstrip("/")

        if request_path.startswith(self.path):
            return request_path.removeprefix(self.path).lstrip("/"), None

        return None


class StaticFilesMiddleware(BaseFilesMiddleware):
    # pylint: disable=too-many-arguments,too-many-positional-arguments
    def __init__(
        self,
        app,
        path: str,
        root: Path,
        filelist: dict[str, os.stat_result],
        mappings: dict[str, str],
    ):
        super().__init__(app, path, root)
        self.filelist = filelist
        self.mappings = mappings

    def get_file_to_serve(self, scope: dict) -> tuple[str, os.stat_result] | None:
        request_path = scope["path"].lstrip("/")

        file_to_serve = None
        if file_path := self.mappings.get(request_path):
            file_to_serve = file_path
        elif request_path.startswith(self.path):
            file_path = os.path.join(self.root, request_path.removeprefix(self.path))
            if file_path in self.filelist:
                file_to_serve = file_path

        if file_to_serve:
            stat_result = self.filelist[file_to_serve]
            return file_to_serve, stat_result

        return None

    async def __call__(self, scope, receive, send):
        try:
            await super().__call__(scope, receive, send)
        except HTTPNotFoundException:
            if scope["path"] == "/":
                new_scope = scope | {"path": f"{self.path}index.html"}
                await super().__call__(new_scope, receive, send)
            else:
                raise



import pytest

from zayt.di.container import Container
from zayt.di.error import ServiceNotFoundError


class Interface:
    pass


class Impl1(Interface):
    pass


class Impl2(Interface):
    pass


class AnotherInterface:
    pass


class AnotherImpl(AnotherInterface):
    pass


def test_iter_service(ioc: Container):
    ioc.register(Impl1, provides=Interface, name="impl1")
    ioc.register(Impl2, provides=Interface, name="impl2")
    ioc.register(AnotherImpl, provides=AnotherInterface)

    result = list(ioc.iter_service(Interface))

    assert result == [(Impl1, "impl1"), (Impl2, "impl2")]


def test_iter_all_services(ioc: Container):
    ioc.register(Impl1, provides=Interface, name="impl1")
    ioc.register(Impl2, provides=Interface, name="impl2")
    ioc.register(AnotherImpl, provides=AnotherInterface)

    result = set(ioc.iter_all_services())
    assert result == {
        (Interface, Impl1, "impl1"),
        (Interface, Impl2, "impl2"),
        (AnotherInterface, AnotherImpl, None)
    }


def test_iter_service_with_not_registered_service_should_fail(ioc: Container):
    with pytest.raises(ServiceNotFoundError):
        list(ioc.iter_service(Interface))



from typing import Annotated

import pytest

from zayt.di.container import Container
from zayt.di.error import ServiceAlreadyRegisteredError
from zayt.di.inject import Inject


class Service1:
    pass


class Service2:
    service1: Annotated[Service1, Inject]


class Service3:
    pass


class Interface:
    pass


class Implementation(Interface):
    pass


def test_has_service(ioc: Container):
    ioc.register(Service1)
    assert ioc.has(Service1)


async def test_service_with_provided_interface(ioc: Container):
    ioc.register(Implementation, provides=Interface)

    instance = await ioc.get(Interface)
    assert isinstance(instance, Implementation)


async def test_inject_singleton(ioc: Container):
    ioc.register(Service1)
    ioc.register(Service2)

    instance = await ioc.get(Service2)
    assert isinstance(instance, Service2)
    assert isinstance(instance.service1, Service1)

    other_instance = await ioc.get(Service2)
    assert other_instance is instance
    assert other_instance.service1 is instance.service1


async def test_interface_and_implementation(ioc: Container):
    ioc.register(Implementation, provides=Interface)

    instance = await ioc.get(Interface)
    assert isinstance(instance, Implementation)


async def test_register_a_service_twice_should_fail(ioc: Container):
    class Implementation2(Interface):
        pass

    ioc.register(Implementation, provides=Interface)
    with pytest.raises(ServiceAlreadyRegisteredError):
        ioc.register(Implementation2, provides=Interface)



from pathlib import Path

import pytest

from zayt.conf.defaults import default_settings
from zayt.conf.settings import (
    Settings,
    SettingsError,
    _merge_recursive,
    get_settings,
    get_settings_for_profile,
)


def test_get_settings(monkeypatch):
    monkeypatch.chdir(Path(__file__).parent / "base")

    result = get_settings()
    assert result == default_settings | {
        "prop": "value",
        "list": [1, 2, 3],
        "dict": Settings({"a": 1, "b": 2, "c": 3}),
    }


@pytest.mark.parametrize(
    "profile",
    ["dev", "stg", "prd"],
)
def test_get_settings_with_profile(monkeypatch, profile):
    monkeypatch.chdir(Path(__file__).parent / "profiles")
    monkeypatch.setenv("ZAYT_PROFILE", profile)

    result = get_settings()
    assert result == default_settings | {
        "name": "application",
        "profile": profile,
    }


@pytest.mark.parametrize(
    ["profile_a", "profile_b"],
    [["dev", "stg"], ["stg", "prd"], ["prd", "dev"]],
)
def test_get_settings_with_multiple_profiles(monkeypatch, profile_a, profile_b):
    monkeypatch.chdir(Path(__file__).parent / "multiple_profiles")
    monkeypatch.setenv("ZAYT_PROFILE", f"{profile_a}, {profile_b}")

    result = get_settings()
    assert result == default_settings | {
        f"profile_{profile_a}": True,
        f"profile_{profile_b}": True,
    }


@pytest.mark.parametrize(
    "profile,expected",
    [
        (None, {"name": "application"}),
        ("dev", {"profile": "dev"}),
        ("stg", {"profile": "stg"}),
        ("prd", {"profile": "prd"}),
    ],
    ids=["None", "dev", "stg", "prd"],
)
def test_get_settings_for_profile(monkeypatch, profile, expected):
    monkeypatch.chdir(Path(__file__).parent / "profiles")

    result, _ = get_settings_for_profile(profile)
    assert result == expected


def test_empty_settings_file(monkeypatch):
    monkeypatch.chdir(Path(__file__).parent / "empty")

    result, _ = get_settings_for_profile()
    assert result == {}


def test_configure_settings_dir(monkeypatch):
    monkeypatch.setenv(
        "ZAYT_SETTINGS_DIR",
        str(Path(__file__).parent / "base" / "configuration"),
    )

    result = get_settings()
    assert result == default_settings | {
        "prop": "value",
        "list": [1, 2, 3],
        "dict": Settings({"a": 1, "b": 2, "c": 3}),
    }


def test_configure_settings_file(monkeypatch):
    monkeypatch.chdir(str(Path(__file__).parent / "alternate"))
    monkeypatch.setenv(
        "ZAYT_SETTINGS_FILE",
        "application.py",
    )

    result = get_settings()
    assert result == default_settings | {
        "prop": "value",
        "list": [1, 2, 3],
        "dict": Settings({"a": 1, "b": 2, "c": 3}),
    }


def test_configure_settings_dir_and_file(monkeypatch):
    monkeypatch.setenv(
        "ZAYT_SETTINGS_DIR",
        str(Path(__file__).parent / "alternate" / "configuration"),
    )
    monkeypatch.setenv(
        "ZAYT_SETTINGS_FILE",
        "application.py",
    )

    result = get_settings()
    assert result == default_settings | {
        "prop": "value",
        "list": [1, 2, 3],
        "dict": Settings({"a": 1, "b": 2, "c": 3}),
    }


def test_configure_settings_file_with_profile(monkeypatch):
    monkeypatch.chdir(str(Path(__file__).parent / "alternate"))
    monkeypatch.setenv(
        "ZAYT_SETTINGS_FILE",
        "application.py",
    )

    monkeypatch.setenv("ZAYT_PROFILE", "prd")

    result = get_settings()
    assert result == default_settings | {
        "profile": "prd",
        "prop": "value",
        "list": [1, 2, 3],
        "dict": Settings({"a": 1, "b": 2, "c": 3}),
    }


@pytest.mark.parametrize(
    "env",
    ["dev", "stg", "prd"],
)
def test_configure_env_setttings(monkeypatch, env):
    monkeypatch.setenv(
        "ZAYT_SETTINGS_DIR",
        str(Path(__file__).parent / "profiles/configuration"),
    )
    monkeypatch.setenv("ZAYT_PROFILE", env)

    result = get_settings()
    assert result == default_settings | {
        "name": "application",
        "profile": env,
    }


@pytest.mark.parametrize(
    "env",
    ["dev", "stg", "prd"],
)
def test_override_settings(monkeypatch, env):
    monkeypatch.chdir(Path(__file__).parent / "override")

    result = get_settings()
    assert result == default_settings | {"value": "base"}

    monkeypatch.setenv("ZAYT_PROFILE", env)

    result = get_settings()
    assert result == default_settings | {"value": env}


def test_settings_class_dict_access(monkeypatch):
    monkeypatch.chdir(Path(__file__).parent / "base")

    settings = get_settings()
    assert settings["prop"] == "value"
    assert settings["list"] == [1, 2, 3]
    assert settings["dict"] == {"a": 1, "b": 2, "c": 3}


@pytest.mark.parametrize(
    "env",
    ["dev", "stg", "prd"],
)
def test_setttings_class_dict_access_env(monkeypatch, env):
    monkeypatch.chdir(Path(__file__).parent / "profiles")
    monkeypatch.setenv("ZAYT_PROFILE", env)

    settings = get_settings()
    assert settings["name"] == "application"
    assert settings["profile"] == env


def test_invalid_settings_object_should_fail(monkeypatch):
    application_path = Path(__file__).parent / "invalid_configuration" / "invalid_yaml"
    monkeypatch.chdir(application_path)

    settings_path = application_path / "configuration" / "settings.py"

    with pytest.raises(
        SettingsError, match=f"cannot load settings from {settings_path}"
    ) as exc:
        get_settings_for_profile()

    assert isinstance(exc.value.__cause__, TypeError)
    assert str(exc.value.__cause__) == "invalid settings object type"


@pytest.mark.parametrize(
    "settings,extra,expected",
    [
        ({"a": 1}, {"b": 2}, {"a": 1, "b": 2}),
        ({"a": 1}, {"a": 2}, {"a": 2}),
        ({"a": {"a": 1}}, {"a": {"b": 2}}, {"a": {"a": 1, "b": 2}}),
        ({"a": {"a": 1}}, {"a": {"a": 2}}, {"a": {"a": 2}}),
        ({"a": {"a": 1}}, {"a": 1}, {"a": 1}),
        ({"a": 1}, {"a": {"a": 1}}, {"a": {"a": 1}}),
    ],
    ids=[
        "add value",
        "replace value",
        "add nested value",
        "replace nested value",
        "replace dict with value",
        "replace value with dict",
    ],
)
def test_merge_recursive(settings, extra, expected):
    _merge_recursive(settings, extra)

    assert settings == expected


def test_settings_item_access():
    settings = Settings({"a": 1})

    assert settings["a"] == 1


def test_settings_nested_item_access():
    settings = Settings({"a": {"b": 1}})

    assert settings["a"]["b"] == 1


def test_settings_attribute_access():
    settings = Settings({"a": 1})

    assert settings.a == 1


def test_settings_nested_attribute_access():
    settings = Settings({"a": {"b": 1}})

    assert settings.a.b == 1


def test_settings_non_existent_attribute_should_fail():
    settings = Settings({})

    with pytest.raises(AttributeError, match="a"):
        settings.a


def test_settings_mixed_item_attribute_access():
    settings = Settings({"a": {"b": 1}})

    assert settings["a"].b == 1
    assert settings.a["b"] == 1


def test_settings_method_access():
    settings = Settings({"a": 1})

    assert settings.get("a") == 1


def test_settings_method_access_default_value():
    settings = Settings({})

    assert settings.get("a") is None
    assert settings.get("a", "b") == "b"


def test_settings_get_nonexistent_item_should_fail():
    settings = Settings({})
    with pytest.raises(KeyError, match="a"):
        settings["a"]


def test_settings_contains_item():
    settings = Settings({"a": 1})
    assert "a" in settings


def test_settings_eq():
    settings1 = Settings({"a": 1})
    settings2 = Settings({"a": 1})
    assert settings1 == settings2


def test_settings_not_eq():
    settings1 = Settings({"a": 1})
    settings2 = Settings({"a": 2})
    assert settings1 != settings2


def test_settings_eq_dict():
    settings = Settings({"a": 1})
    other = {"a": 1}
    assert settings == other


def test_settings_not_eq_dict():
    settings = Settings({"a": 1})
    other = {"a": 2}
    assert settings != other


def test_settings_not_eq_object():
    settings = Settings({})
    assert settings != "settings"


def test_settings_str():
    data = {"a": 1}
    settings = Settings(data)
    assert str(settings) == str(data)


def test_settings_repr():
    data = {"a": 1}
    settings = Settings(data)
    assert repr(settings) == repr(data)



import os
from importlib.util import find_spec
from urllib.parse import urlparse

import pytest

from zayt.conf.defaults import default_settings
from zayt.conf.settings import Settings
from zayt.ext.redis.service import make_service

REDIS_URL = os.getenv("REDIS_URL")

pytestmark = [
    pytest.mark.skipif(REDIS_URL is None, reason="REDIS_URL not defined"),
    pytest.mark.skipif(find_spec("redis") is None, reason="redis not present"),
]

PARSED_URL = urlparse(REDIS_URL) if REDIS_URL else None


async def _test_engine_service(settings: Settings):
    service = make_service("default")(settings)
    async for redis in service:
        result = await redis.ping()
        assert result


async def test_make_service_with_url():
    settings = Settings(
        default_settings
        | {
            "redis": {
                "default": {
                    "url": REDIS_URL,
                },
            },
        }
    )

    await _test_engine_service(settings)


@pytest.mark.skipif(
    PARSED_URL is not None and PARSED_URL.password is None,
    reason="url without password",
)
async def test_make_service_with_url_username_password():
    settings = Settings(
        default_settings
        | {
            "redis": {
                "default": {
                    "url": f"redis://{PARSED_URL.hostname}:{PARSED_URL.port}{PARSED_URL.path}",
                    "username": PARSED_URL.username,
                    "password": PARSED_URL.password,
                },
            },
        }
    )

    await _test_engine_service(settings)


async def test_make_service_with_url_components():
    params = {
        "host": PARSED_URL.hostname,
        "port": PARSED_URL.port,
        "db": int(PARSED_URL.path.strip("/")),
    }

    if (username := PARSED_URL.username) and (password := PARSED_URL.password):
        params |= {
            "username": username,
            "password": password,
        }

    settings = Settings(
        default_settings
        | {
            "redis": {"default": params},
        }
    )

    await _test_engine_service(settings)


async def test_make_service_with_options():
    settings = Settings(
        default_settings
        | {
            "redis": {
                "default": {
                    "url": REDIS_URL,
                    "client_name": "zayt",
                },
            },
        }
    )

    service = make_service("default")(settings)
    async for redis in service:
        connection = await redis.connection_pool.get_connection()
        assert connection.client_name == "zayt"


async def test_make_service_with_retry():
    # pylint: disable=import-outside-toplevel
    # pylint: disable=import-outside-toplevel
    from redis.backoff import ConstantBackoff
    from redis.retry import Retry

    settings = Settings(
        default_settings
        | {
            "redis": {
                "default": {
                    "url": REDIS_URL,
                    "retry": Retry(retries=1, backoff=ConstantBackoff(1.0)),
                },
            },
        }
    )

    service = make_service("default")(settings)
    async for redis in service:
        connection = await redis.connection_pool.get_connection()
        assert connection.retry is not None



import copy

import pytest

from zayt.conf.defaults import default_settings
from zayt.conf.settings import Settings
from zayt.web.application import Zayt


async def test_non_existent_module_should_fail():
    settings = Settings(
        copy.copy(default_settings)
        | {
            "modules": [
                f"{__package__}.application",
                "does.not.exist",
            ],
        }
    )

    with pytest.raises(ModuleNotFoundError):
        Zayt(settings)



from http import HTTPMethod

import pytest

from zayt.web.routing.decorator import (
    ATTRIBUTE_HANDLER,
    ATTRIBUTE_WEBSOCKET,
    HandlerInfo,
    WebSocketInfo,
    delete,
    get,
    patch,
    post,
    put,
    route,
    websocket,
)
from zayt.web.routing.exception import (
    HandlerMissingRequiredArgumentError,
    HandlerNotAsyncError,
    HandlerRequestTypeError,
    HandlerUntypedParametersError,
)


@pytest.mark.parametrize(
    "decorator,method",
    [
        (get, HTTPMethod.GET),
        (post, HTTPMethod.POST),
        (put, HTTPMethod.PUT),
        (patch, HTTPMethod.PATCH),
        (delete, HTTPMethod.DELETE),
    ],
    ids=["get", "post", "put", "patch", "delete"],
)
def test_route_decorator_without_path(decorator, method):
    async def handler(_request, _response):
        pass

    decorator(handler)

    assert hasattr(handler, ATTRIBUTE_HANDLER)
    assert getattr(handler, ATTRIBUTE_HANDLER) == HandlerInfo({(method, "")})


def test_websocket_decorator_without_path():
    async def handler(_websocket):
        pass

    websocket(handler)

    assert hasattr(handler, ATTRIBUTE_WEBSOCKET)
    assert getattr(handler, ATTRIBUTE_WEBSOCKET) == WebSocketInfo({""})


@pytest.mark.parametrize(
    "decorator,method",
    [
        (get, HTTPMethod.GET),
        (post, HTTPMethod.POST),
        (put, HTTPMethod.PUT),
        (patch, HTTPMethod.PATCH),
        (delete, HTTPMethod.DELETE),
    ],
    ids=["get", "post", "put", "patch", "delete"],
)
def test_route_decorator_with_path(decorator, method):
    async def handler(_request, _response):
        pass

    decorator("path")(handler)

    assert hasattr(handler, ATTRIBUTE_HANDLER)
    assert getattr(handler, ATTRIBUTE_HANDLER) == HandlerInfo({(method, "path")})


def test_websocket_decorator_with_path():
    async def handler(_websocket):
        pass

    websocket("path")(handler)

    assert hasattr(handler, ATTRIBUTE_WEBSOCKET)
    assert getattr(handler, ATTRIBUTE_WEBSOCKET) == WebSocketInfo({"path"})


def test_handler_with_less_than_one_parameters_should_fail():
    async def func0():
        pass

    with pytest.raises(HandlerMissingRequiredArgumentError):
        route(HTTPMethod.GET, "")(func0)


def test_websocket_handler_with_less_than_one_parameter_should_fail():
    async def func():
        pass

    with pytest.raises(HandlerMissingRequiredArgumentError):
        websocket("")(func)


def test_handler_with_wrong_request_parameter_should_fail():
    async def handler(_request: str, _response):
        pass

    with pytest.raises(HandlerRequestTypeError):
        route(HTTPMethod.GET, "")(handler)


def test_websocket_handler_with_wrong_parameter_should_fail():
    async def handler(_websocket: str):
        pass

    with pytest.raises(HandlerRequestTypeError):
        websocket("")(handler)


def test_handler_with_untyped_parameters_should_fail():
    async def handler(_request, _response, _untyped):
        pass

    with pytest.raises(HandlerUntypedParametersError):
        route(HTTPMethod.GET, "")(handler)


def test_non_async_handler_should_fail():
    def handler(_request, _response):
        pass

    with pytest.raises(HandlerNotAsyncError):
        route(HTTPMethod.GET, "")(handler)


def test_multiple_handler_decorators():
    async def handler(_request, _response):
        pass

    get("path")(handler)
    post("another")(handler)

    assert getattr(handler, ATTRIBUTE_HANDLER) == HandlerInfo(
        {
            (HTTPMethod.GET, "path"),
            (HTTPMethod.POST, "another"),
        }
    )


def test_multiple_websocket_decorators():
    async def handler(_websocket):
        pass

    websocket("path")(handler)
    websocket("another")(handler)

    assert getattr(handler, ATTRIBUTE_WEBSOCKET) == WebSocketInfo({"path", "another"})



import asyncio
import multiprocessing as mp
from pathlib import Path

from httpx import ASGITransport, AsyncClient

from .util import with_app


@with_app
async def _run(app):
    async with AsyncClient(
        transport=ASGITransport(app=app), base_url="http://server"
    ) as client:
        response = await client.get("/")
        assert response.status_code == 200
        assert (
            '<img src="/static/python-powered.png" alt="python powered" />'
            in response.text
        )

        response = await client.get("/favicon.ico")
        assert response.headers.get("content-type") == "image/x-icon"

        response = await client.get("/static/python-powered.png")
        assert response.headers.get("content-type") == "image/png"


def run():
    return asyncio.run(_run())


def test_example(monkeypatch):
    path = Path(__file__).parents[2] / "examples" / "middleware_files"
    monkeypatch.syspath_prepend(path)
    monkeypatch.chdir(path)

    with mp.Pool(processes=1) as pool:
        pool.apply(run)



from collections.abc import Awaitable, Callable
from typing import NamedTuple, TypeAlias

from asgikit import Request

ATTRIBUTE_EXCEPTION_HANDLER = "__zayt_exception_handler__"

ExceptionHandlerType: TypeAlias = Callable[[Request, BaseException, ...], Awaitable]


class ExceptionHandlerInfo(NamedTuple):
    exception_class: type[BaseException]


def exception_handler(exc: type[BaseException]):
    assert issubclass(exc, BaseException)

    def inner(handler: ExceptionHandlerType):
        setattr(
            handler,
            ATTRIBUTE_EXCEPTION_HANDLER,
            ExceptionHandlerInfo(exception_class=exc),
        )
        return handler

    return inner



import inspect
import typing
from collections.abc import Callable
from typing import Annotated, TypeVar, dataclass_transform

from zayt.di.inject import Inject
from zayt.di.service.model import InjectableType, ServiceInfo

__all__ = ("service", "ATTRIBUTE_DI_SERVICE")

T = TypeVar("T")

ATTRIBUTE_DI_SERVICE = "__zayt_di_service__"


def _is_inject(value) -> bool:
    origin = typing.get_origin(value)
    if not origin or origin is not Annotated:
        return False

    # if origin is Annotated, args has always at least 2 values
    args = typing.get_args(value)

    return isinstance(args[1], Inject) or args[1] is Inject


def _service(
    injectable: InjectableType,
    attribute_name: str,
    attribute_value,
) -> InjectableType:
    setattr(injectable, attribute_name, attribute_value)

    if inspect.isclass(injectable):
        dependencies = [
            dependency
            for dependency, annotation in inspect.get_annotations(injectable).items()
            if _is_inject(annotation)
        ]

        # save a reference to the original constructor
        original_init = getattr(injectable, "__init__", None)

        def init(self, *args, **kwargs):
            """Generated init method for service

            Positional and keyword arguments will be set to declared dependencies.
            Dependencies without an argument to set their value will be None.
            Remaining arguments will be ignored.
            """

            # call original constructor
            if original_init:
                original_init(self)

            positional_params = [d for d in dependencies if d not in kwargs]

            # set positional argument values
            values = dict(zip(positional_params, args))

            # set keyword argument values
            values |= {k: v for k, v in kwargs.items() if k in dependencies}

            # the rest of the dependencies will be set to None
            param_keys = values.keys()
            values |= {d: None for d in dependencies if d not in param_keys}

            for k, v in values.items():
                setattr(self, k, v)

        setattr(injectable, "__init__", init)

    return injectable


@dataclass_transform(eq_default=False)
def service(
    injectable: T = None,
    /,
    *,
    provides: type = None,
    name: str = None,
    scoped=False,
) -> T | Callable[[T], T]:
    """Declare a class or function as a service

    For classes, a constructor will be generated to help create instances
    outside the dependency injection context

    :param injectable: The class or function defining the service
    :param provides: The interface that this service provides
    :param name: The name of the service
    :param scoped: If True, the service will be scoped to a context object, usually the web Request
    """

    def inner(inner_injectable) -> T:
        return _service(
            inner_injectable, ATTRIBUTE_DI_SERVICE, ServiceInfo(provides, name, scoped)
        )

    return inner(injectable) if injectable else inner



import uuid

from structlog.contextvars import bind_contextvars, unbind_contextvars

from zayt.util.asgi import get_header


async def request_id_middleware(app):
    async def handler(scope, receive, send):
        request_id = get_header("x-request-id", scope) or str(uuid.uuid4())

        try:
            bind_contextvars(request_id=request_id)
            await app(scope, receive, send)
        finally:
            unbind_contextvars("request_id")

    return handler
